import os
import shutil
import tempfile
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast, overload

from uipath.core.tracing import traced

from ..common._base_service import BaseService
from ..common._bindings import resource_override
from ..common._config import UiPathApiConfig
from ..common._execution_context import UiPathExecutionContext
from ..common._folder_context import FolderContext, header_folder
from ..common._models import Endpoint, RequestSpec
from ..common.constants import TEMP_ATTACHMENTS_FOLDER
from ..common.paging import PagedResult
from ..common.validation import validate_pagination_params
from ..errors import EnrichedException
from ._attachments_service import AttachmentsService
from .job import Job


class JobsService(FolderContext, BaseService):
    """Service for managing API payloads and job inbox interactions.

    A job represents a single execution of an automation - it is created when you start
      a process and contains information about that specific run, including its status,
      start time, and any input/output data.
    """

    # Pagination limits
    MAX_PAGE_SIZE = 1000  # Maximum items per page
    MAX_SKIP_OFFSET = 10000  # Maximum skip offset

    def __init__(
        self, config: UiPathApiConfig, execution_context: UiPathExecutionContext
    ) -> None:
        super().__init__(config=config, execution_context=execution_context)
        self._attachments_service = AttachmentsService(config, execution_context)
        # Define the temp directory path for local attachments
        self._temp_dir = os.path.join(tempfile.gettempdir(), TEMP_ATTACHMENTS_FOLDER)
        os.makedirs(self._temp_dir, exist_ok=True)

    @overload
    def resume(self, *, inbox_id: str, payload: Any) -> None: ...

    @overload
    def resume(self, *, job_id: str, payload: Any) -> None: ...

    @traced(name="jobs_resume", run_type="uipath")
    def resume(
        self,
        *,
        inbox_id: Optional[str] = None,
        job_id: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
        payload: Any,
    ) -> None:
        """Sends a payload to resume a paused job waiting for input, identified by its inbox ID.

        Args:
            inbox_id (Optional[str]): The inbox ID of the job.
            job_id (Optional[str]): The job ID of the job.
            folder_key (Optional[str]): The key of the folder to execute the process in. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder to execute the process in. Override the default one set in the SDK config.
            payload (Any): The payload to deliver.
        """
        if job_id is None and inbox_id is None:
            raise ValueError("Either job_id or inbox_id must be provided")

        # for type checking
        job_id = str(job_id)
        inbox_id = (
            inbox_id
            if inbox_id
            else self._retrieve_inbox_id(
                job_id=job_id,
                folder_key=folder_key,
                folder_path=folder_path,
            )
        )
        spec = self._resume_spec(
            inbox_id=inbox_id,
            payload=payload,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        self.request(
            spec.method,
            url=spec.endpoint,
            headers=spec.headers,
            json=spec.json,
        )

    async def resume_async(
        self,
        *,
        inbox_id: Optional[str] = None,
        job_id: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
        payload: Any,
    ) -> None:
        """Asynchronously sends a payload to resume a paused job waiting for input, identified by its inbox ID.

        Args:
            inbox_id (Optional[str]): The inbox ID of the job. If not provided, the execution context will be used to retrieve the inbox ID.
            job_id (Optional[str]): The job ID of the job.
            folder_key (Optional[str]): The key of the folder to execute the process in. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder to execute the process in. Override the default one set in the SDK config.
            payload (Any): The payload to deliver.

        Examples:
            ```python
            import asyncio

            from uipath.platform import UiPath

            sdk = UiPath()


            async def main():  # noqa: D103
                payload = await sdk.jobs.resume_async(job_id="38073051", payload="The response")

            asyncio.run(main())
            ```
        """
        if job_id is None and inbox_id is None:
            raise ValueError("Either job_id or inbox_id must be provided")

        # for type checking
        job_id = str(job_id)
        inbox_id = (
            inbox_id
            if inbox_id
            else self._retrieve_inbox_id(
                job_id=job_id,
                folder_key=folder_key,
                folder_path=folder_path,
            )
        )

        spec = self._resume_spec(
            inbox_id=inbox_id,
            payload=payload,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        await self.request_async(
            spec.method,
            url=spec.endpoint,
            headers=spec.headers,
            json=spec.json,
        )

    @property
    def custom_headers(self) -> Dict[str, str]:
        return self.folder_headers

    @traced(name="jobs_list", run_type="uipath")
    def list(
        self,
        *,
        folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
        filter: Optional[str] = None,
        orderby: Optional[str] = None,
        skip: int = 0,
        top: int = 100,
    ) -> PagedResult[Job]:
        """List jobs using OData API with offset-based pagination.

        Returns a single page of results with pagination metadata.

        Args:
            folder_path: Folder path to filter jobs
            folder_key: Folder key (mutually exclusive with folder_path)
            filter: OData $filter expression (e.g., "State eq 'Successful'")
            orderby: OData $orderby expression (e.g., "CreationTime desc")
            skip: Number of items to skip (default 0, max 10000)
            top: Maximum items per page (default 100, max 1000)

        Returns:
            PagedResult[Job]: Page of jobs with pagination metadata

        Raises:
            ValueError: If skip or top parameters are invalid

        Examples:
            >>> result = sdk.jobs.list(top=100)
            >>> for job in result.items:
            ...     print(job.key, job.state)
            >>> print(f"Has more: {result.has_more}")
        """
        validate_pagination_params(
            skip=skip,
            top=top,
            max_skip=self.MAX_SKIP_OFFSET,
            max_top=self.MAX_PAGE_SIZE,
        )

        spec = self._list_spec(
            folder_path=folder_path,
            folder_key=folder_key,
            filter=filter,
            orderby=orderby,
            skip=skip,
            top=top,
        )
        response = self.request(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            headers=spec.headers,
        ).json()

        items = response.get("value", [])
        jobs = [Job.model_validate(item) for item in items]

        return PagedResult(
            items=jobs,
            has_more=len(items) == top,
            skip=skip,
            top=top,
        )

    @traced(name="jobs_list", run_type="uipath")
    async def list_async(
        self,
        *,
        folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
        filter: Optional[str] = None,
        orderby: Optional[str] = None,
        skip: int = 0,
        top: int = 100,
    ) -> PagedResult[Job]:
        """Async version of list() with offset-based pagination."""
        validate_pagination_params(
            skip=skip,
            top=top,
            max_skip=self.MAX_SKIP_OFFSET,
            max_top=self.MAX_PAGE_SIZE,
        )

        spec = self._list_spec(
            folder_path=folder_path,
            folder_key=folder_key,
            filter=filter,
            orderby=orderby,
            skip=skip,
            top=top,
        )
        response = (
            await self.request_async(
                spec.method,
                url=spec.endpoint,
                params=spec.params,
                headers=spec.headers,
            )
        ).json()

        items = response.get("value", [])
        jobs = [Job.model_validate(item) for item in items]

        return PagedResult(
            items=jobs,
            has_more=len(items) == top,
            skip=skip,
            top=top,
        )

    @traced(name="jobs_stop", run_type="uipath")
    def stop(
        self,
        *,
        job_keys: List[str],
        strategy: str = "SoftStop",
        folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
    ) -> None:
        """Stop one or more jobs with specified strategy.

        This method uses bulk resolution to efficiently stop multiple jobs,
        preventing N+1 query issues. Requests are automatically chunked for
        large job lists to avoid URL length constraints.

        Args:
            job_keys: List of job UUID keys to stop
            strategy: Stop strategy - "SoftStop" (graceful) or "Kill" (force)
            folder_path: Folder path
            folder_key: Folder key

        Raises:
            ValueError: If any job key is not a valid UUID format
            LookupError: If any job keys are not found

        Examples:
            >>> sdk.jobs.stop(job_keys=["ee9327fd-237d-419e-86ef-9946b34461e3"])
            >>> sdk.jobs.stop(job_keys=["key1", "key2"], strategy="Kill")

        Note:
            Large batches are automatically chunked (50 keys per request) to
            avoid URL length limits. The method supports stopping hundreds of
            jobs efficiently.
        """
        job_ids = self._resolve_job_identifiers(
            job_keys=job_keys,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        spec = self._stop_spec(
            job_ids=job_ids,
            strategy=strategy,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        self.request(
            spec.method,
            url=spec.endpoint,
            json=spec.json,
            headers=spec.headers,
        )

    @traced(name="jobs_stop", run_type="uipath")
    async def stop_async(
        self,
        *,
        job_keys: List[str],
        strategy: str = "SoftStop",
        folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
    ) -> None:
        """Async version of stop()."""
        job_ids = await self._resolve_job_identifiers_async(
            job_keys=job_keys,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        spec = self._stop_spec(
            job_ids=job_ids,
            strategy=strategy,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        await self.request_async(
            spec.method,
            url=spec.endpoint,
            json=spec.json,
            headers=spec.headers,
        )

    @traced(name="jobs_restart", run_type="uipath")
    def restart(
        self,
        *,
        job_key: str,
        folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
    ) -> Job:
        """Restart a completed or failed job.

        Args:
            job_key: Job UUID key to restart
            folder_path: Folder path
            folder_key: Folder key

        Returns:
            Job: The restarted job

        Examples:
            >>> restarted_job = sdk.jobs.restart(job_key="ee9327fd-237d-419e-86ef-9946b34461e3")
            >>> print(restarted_job.state)
        """
        job_ids = self._resolve_job_identifiers(
            job_keys=[job_key],
            folder_key=folder_key,
            folder_path=folder_path,
        )

        spec = self._restart_spec(
            job_id=job_ids[0],
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = self.request(
            spec.method,
            url=spec.endpoint,
            json=spec.json,
            headers=spec.headers,
        )

        return Job.model_validate(response.json())

    @traced(name="jobs_restart", run_type="uipath")
    async def restart_async(
        self,
        *,
        job_key: str,
        folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
    ) -> Job:
        """Async version of restart()."""
        job_ids = await self._resolve_job_identifiers_async(
            job_keys=[job_key],
            folder_key=folder_key,
            folder_path=folder_path,
        )

        spec = self._restart_spec(
            job_id=job_ids[0],
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = await self.request_async(
            spec.method,
            url=spec.endpoint,
            json=spec.json,
            headers=spec.headers,
        )

        return Job.model_validate(response.json())

    @traced(name="jobs_exists", run_type="uipath")
    def exists(
        self,
        job_key: str,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> bool:
        """Check if job exists.

        Args:
            job_key: Job UUID key
            folder_key: Folder key
            folder_path: Folder path

        Returns:
            bool: True if job exists, False otherwise

        Examples:
            >>> if sdk.jobs.exists(job_key="ee9327fd-237d-419e-86ef-9946b34461e3"):
            ...     print("Job found")
        """
        try:
            self.retrieve(
                job_key=job_key, folder_key=folder_key, folder_path=folder_path
            )
            return True
        except LookupError:
            return False

    @traced(name="jobs_exists", run_type="uipath")
    async def exists_async(
        self,
        job_key: str,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> bool:
        """Async version of exists()."""
        try:
            await self.retrieve_async(
                job_key=job_key, folder_key=folder_key, folder_path=folder_path
            )
            return True
        except LookupError:
            return False

    @resource_override(resource_type="process", resource_identifier="process_name")
    @traced(name="jobs_retrieve", run_type="uipath")
    def retrieve(
        self,
        job_key: str,
        *,
        folder_key: str | None = None,
        folder_path: str | None = None,
        process_name: str | None = None,
    ) -> Job:
        """Retrieve a job identified by its key.

        Args:
            job_key (str): The job unique identifier.
            folder_key (Optional[str]): The key of the folder in which the job was executed.
            folder_path (Optional[str]): The path of the folder in which the job was executed.
            process_name: process name hint for resource override

        Returns:
            Job: The retrieved job.

        Raises:
            LookupError: If the job with the specified key is not found.

        Examples:
            ```python
            from uipath.platform import UiPath

            sdk = UiPath()
            job = sdk.jobs.retrieve(job_key="ee9327fd-237d-419e-86ef-9946b34461e3", folder_path="Shared")
            ```
        """
        spec = self._retrieve_spec(
            job_key=job_key, folder_key=folder_key, folder_path=folder_path
        )
        try:
            response = self.request(
                spec.method,
                url=spec.endpoint,
                headers=spec.headers,
            )
            return Job.model_validate(response.json())
        except EnrichedException as e:
            if e.status_code == 404:
                raise LookupError(f"Job with key '{job_key}' not found") from e
            raise

    @resource_override(resource_type="process", resource_identifier="process_name")
    @traced(name="jobs_retrieve_async", run_type="uipath")
    async def retrieve_async(
        self,
        job_key: str,
        *,
        folder_key: str | None = None,
        folder_path: str | None = None,
        process_name: str | None = None,
    ) -> Job:
        """Asynchronously retrieve a job identified by its key.

        Args:
            job_key (str): The job unique identifier.
            folder_key (Optional[str]): The key of the folder in which the job was executed.
            folder_path (Optional[str]): The path of the folder in which the job was executed.
            process_name: process name hint for resource override

        Returns:
            Job: The retrieved job.

        Raises:
            LookupError: If the job with the specified key is not found.

        Examples:
            ```python
            import asyncio

            from uipath.platform import UiPath

            sdk = UiPath()


            async def main():  # noqa: D103
                job = await sdk.jobs.retrieve_async(job_key="ee9327fd-237d-419e-86ef-9946b34461e3", folder_path="Shared")

            asyncio.run(main())
            ```
        """
        spec = self._retrieve_spec(
            job_key=job_key, folder_key=folder_key, folder_path=folder_path
        )
        try:
            response = await self.request_async(
                spec.method,
                url=spec.endpoint,
                headers=spec.headers,
            )
            return Job.model_validate(response.json())
        except EnrichedException as e:
            if e.status_code == 404:
                raise LookupError(f"Job with key '{job_key}' not found") from e
            raise

    def _retrieve_inbox_id(
        self,
        *,
        job_id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> str:
        spec = self._retrieve_inbox_id_spec(
            job_id=job_id,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        response = self.request(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            headers=spec.headers,
        )

        response = response.json()
        return self._extract_first_inbox_id(response)

    async def _retrieve_inbox_id_async(
        self,
        *,
        job_id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> str:
        spec = self._retrieve_inbox_id_spec(
            job_id=job_id,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        response = await self.request_async(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            headers=spec.headers,
        )

        response = response.json()
        return self._extract_first_inbox_id(response)

    def retrieve_api_payload(self, inbox_id: str) -> Any:
        """Fetch payload data for API triggers.

        Args:
            inbox_id: The Id of the inbox to fetch the payload for.

        Returns:
            The value field from the API response payload.
        """
        spec = self._retrieve_api_payload_spec(inbox_id=inbox_id)

        response = self.request(
            spec.method,
            url=spec.endpoint,
            headers=spec.headers,
        )

        data = response.json()
        return data.get("payload")

    async def retrieve_api_payload_async(self, inbox_id: str) -> Any:
        """Asynchronously fetch payload data for API triggers.

        Args:
            inbox_id: The Id of the inbox to fetch the payload for.

        Returns:
            The value field from the API response payload.
        """
        spec = self._retrieve_api_payload_spec(inbox_id=inbox_id)

        response = await self.request_async(
            spec.method,
            url=spec.endpoint,
            headers=spec.headers,
        )

        data = response.json()
        return data.get("payload")

    def _retrieve_api_payload_spec(
        self,
        *,
        inbox_id: str,
    ) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint(f"/orchestrator_/api/JobTriggers/GetPayload/{inbox_id}"),
            headers={
                **self.folder_headers,
            },
        )

    def _extract_first_inbox_id(self, response: Any) -> str:
        if len(response["value"]) > 0:
            return response["value"][0]["ItemKey"]
        else:
            raise LookupError("No inbox found")

    def _retrieve_inbox_id_spec(
        self,
        *,
        job_id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint("/orchestrator_/odata/JobTriggers"),
            params={
                "$filter": f"JobId eq {job_id}",
                "$top": 1,
                "$select": "ItemKey",
            },
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def extract_output(self, job: Job) -> Optional[str]:
        """Get the actual output data, downloading from attachment if necessary.

        Args:
            job: The job instance to fetch output data from.

        Returns:
            Parsed output arguments as dictionary, or None if no output
        """
        if job.output_file:
            # Large output stored as attachment
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_path = Path(temp_dir) / f"output_{job.output_file}"
                self._attachments_service.download(
                    key=uuid.UUID(job.output_file), destination_path=temp_path
                )
                with open(temp_path, "r", encoding="utf-8") as f:
                    return f.read()
        elif job.output_arguments:
            # Small output stored inline
            return job.output_arguments
        else:
            return None

    async def extract_output_async(self, job: Job) -> Optional[str]:
        """Asynchronously fetch the actual output data, downloading from attachment if necessary.

        Args:
            job: The job instance to fetch output data from.

        Returns:
            Parsed output arguments as dictionary, or None if no output
        """
        if job.output_file:
            # Large output stored as attachment
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_path = Path(temp_dir) / f"output_{job.output_file}"
                await self._attachments_service.download_async(
                    key=uuid.UUID(job.output_file), destination_path=temp_path
                )
                with open(temp_path, "r", encoding="utf-8") as f:
                    return f.read()
        elif job.output_arguments:
            # Small output stored inline
            return job.output_arguments
        else:
            return None

    def _resume_spec(
        self,
        *,
        inbox_id: str,
        payload: Any = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                f"/orchestrator_/api/JobTriggers/DeliverPayload/{inbox_id}"
            ),
            json={"payload": payload},
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _retrieve_spec(
        self,
        *,
        job_key: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint(
                f"/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.GetByKey(identifier={job_key})"
            ),
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    @traced(name="jobs_list_attachments", run_type="uipath")
    def list_attachments(
        self,
        *,
        job_key: uuid.UUID,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> List[str]:
        """List attachments associated with a specific job.

        This method retrieves all attachments linked to a job by its key.

        Args:
            job_key (uuid.UUID): The key of the job to retrieve attachments for.
            folder_key (Optional[str]): The key of the folder. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder. Override the default one set in the SDK config.

        Returns:
            List[str]: A list of attachment IDs associated with the job.

        Raises:
            Exception: If the retrieval fails.
        """
        spec = self._list_job_attachments_spec(
            job_key=job_key,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = self.request(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            headers=spec.headers,
        ).json()

        return [item.get("attachmentId") for item in response]

    @traced(name="jobs_list_attachments", run_type="uipath")
    async def list_attachments_async(
        self,
        *,
        job_key: uuid.UUID,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> List[str]:
        """List attachments associated with a specific job asynchronously.

        This method asynchronously retrieves all attachments linked to a job by its key.

        Args:
            job_key (uuid.UUID): The key of the job to retrieve attachments for.
            folder_key (Optional[str]): The key of the folder. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder. Override the default one set in the SDK config.

        Returns:
            List[str]: A list of attachment IDs associated with the job.

        Raises:
            Exception: If the retrieval fails.

        Examples:
            ```python
            import asyncio
            from uipath.platform import UiPath

            client = UiPath()

            async def main():
                attachments = await client.jobs.list_attachments_async(
                    job_key=uuid.UUID("123e4567-e89b-12d3-a456-426614174000")
                )
                for attachment_id in attachments:
                    print(f"Attachment ID: {attachment_id}")
            ```
        """
        spec = self._list_job_attachments_spec(
            job_key=job_key,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = (
            await self.request_async(
                spec.method,
                url=spec.endpoint,
                params=spec.params,
                headers=spec.headers,
            )
        ).json()

        return [item.get("attachmentId") for item in response]

    @traced(name="jobs_link_attachment", run_type="uipath")
    def link_attachment(
        self,
        *,
        attachment_key: uuid.UUID,
        job_key: uuid.UUID,
        category: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ):
        """Link an attachment to a job.

        This method links an existing attachment to a specific job.

        Args:
            attachment_key (uuid.UUID): The key of the attachment to link.
            job_key (uuid.UUID): The key of the job to link the attachment to.
            category (Optional[str]): Optional category for the attachment in the context of this job.
            folder_key (Optional[str]): The key of the folder. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder. Override the default one set in the SDK config.

        Raises:
            Exception: If the link operation fails.
        """
        spec = self._link_job_attachment_spec(
            attachment_key=attachment_key,
            job_key=job_key,
            category=category,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        self.request(
            spec.method,
            url=spec.endpoint,
            headers=spec.headers,
            json=spec.json,
        )

    @traced(name="jobs_link_attachment", run_type="uipath")
    async def link_attachment_async(
        self,
        *,
        attachment_key: uuid.UUID,
        job_key: uuid.UUID,
        category: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ):
        """Link an attachment to a job asynchronously.

        This method asynchronously links an existing attachment to a specific job.

        Args:
            attachment_key (uuid.UUID): The key of the attachment to link.
            job_key (uuid.UUID): The key of the job to link the attachment to.
            category (Optional[str]): Optional category for the attachment in the context of this job.
            folder_key (Optional[str]): The key of the folder. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder. Override the default one set in the SDK config.

        Raises:
            Exception: If the link operation fails.
        """
        spec = self._link_job_attachment_spec(
            attachment_key=attachment_key,
            job_key=job_key,
            category=category,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        await self.request_async(
            spec.method,
            url=spec.endpoint,
            headers=spec.headers,
            json=spec.json,
        )

    def _list_job_attachments_spec(
        self,
        job_key: uuid.UUID,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint("/orchestrator_/api/JobAttachments/GetByJobKey"),
            params={
                "jobKey": job_key,
            },
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _link_job_attachment_spec(
        self,
        attachment_key: uuid.UUID,
        job_key: uuid.UUID,
        category: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        return RequestSpec(
            method="POST",
            endpoint=Endpoint("/orchestrator_/api/JobAttachments/Post"),
            json={
                "attachmentId": str(attachment_key),
                "jobKey": str(job_key),
                "category": category,
            },
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    @traced(name="jobs_create_attachment", run_type="uipath")
    def create_attachment(
        self,
        *,
        name: str,
        content: Optional[Union[str, bytes]] = None,
        source_path: Optional[Union[str, Path]] = None,
        job_key: Optional[Union[str, uuid.UUID]] = None,
        category: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> uuid.UUID:
        """Create and upload an attachment, optionally linking it to a job.

        This method handles creating an attachment from a file or memory data.
        If a job key is provided or available in the execution context, the attachment
        will be created in UiPath and linked to the job. If no job is available,
        the file will be saved to a temporary storage folder.

        Note:
            The local storage functionality (when no job is available) is intended for
            local development and debugging purposes only.

        Args:
            name (str): The name of the attachment file.
            content (Optional[Union[str, bytes]]): The content to upload (string or bytes).
            source_path (Optional[Union[str, Path]]): The local path of the file to upload.
            job_key (Optional[Union[str, uuid.UUID]]): The key of the job to link the attachment to.
            category (Optional[str]): Optional category for the attachment in the context of the job.
            folder_key (Optional[str]): The key of the folder. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder. Override the default one set in the SDK config.

        Returns:
            uuid.UUID: The unique identifier for the created attachment, regardless of whether it was
                uploaded to UiPath or stored locally.

        Raises:
            ValueError: If neither content nor source_path is provided, or if both are provided.
            Exception: If the upload fails.

        Examples:
            ```python
            from uipath.platform import UiPath

            client = UiPath()

            # Create attachment from file and link to job
            attachment_id = client.jobs.create_attachment(
                name="document.pdf",
                source_path="path/to/local/document.pdf",
                job_key="38073051"
            )
            print(f"Created and linked attachment: {attachment_id}")

            # Create attachment from memory content (no job available - saves to temp storage)
            attachment_id = client.jobs.create_attachment(
                name="report.txt",
                content="This is a text report"
            )
            print(f"Created attachment: {attachment_id}")
            ```
        """
        # Validate input parameters
        if not (content or source_path):
            raise ValueError("Content or source_path is required")
        if content and source_path:
            raise ValueError("Content and source_path are mutually exclusive")

        # Get job key from context if not explicitly provided
        context_job_key = None
        if job_key is None:
            try:
                context_job_key = self._execution_context.instance_key
            except ValueError:
                # Instance key is not set in environment
                context_job_key = None

        # Check if a job is available
        if job_key is not None or context_job_key is not None:
            # Job is available - create attachment in UiPath and link to job
            actual_job_key = job_key if job_key is not None else context_job_key

            # Create the attachment using the attachments service
            if content is not None:
                attachment_key = self._attachments_service.upload(
                    name=name,
                    content=content,
                    folder_key=folder_key,
                    folder_path=folder_path,
                )
            else:
                # source_path must be provided due to validation check above
                attachment_key = self._attachments_service.upload(
                    name=name,
                    source_path=cast(str, source_path),
                    folder_key=folder_key,
                    folder_path=folder_path,
                )

            # Convert to UUID if string
            if isinstance(actual_job_key, str):
                actual_job_key = uuid.UUID(actual_job_key)

            # Link attachment to job
            self.link_attachment(
                attachment_key=attachment_key,
                job_key=cast(uuid.UUID, actual_job_key),
                category=category,
                folder_key=folder_key,
                folder_path=folder_path,
            )

            return attachment_key
        else:
            # No job available - save to temp folder
            # Generate a UUID to use as identifier
            attachment_id = uuid.uuid4()

            # Create destination file path
            dest_path = os.path.join(self._temp_dir, f"{attachment_id}_{name}")

            # If we have source_path, copy the file
            if source_path is not None:
                source_path_str = (
                    source_path if isinstance(source_path, str) else str(source_path)
                )
                shutil.copy2(source_path_str, dest_path)
            # If we have content, write it to a file
            elif content is not None:
                # Convert string to bytes if needed
                if isinstance(content, str):
                    content = content.encode("utf-8")

                with open(dest_path, "wb") as f:
                    f.write(content)

            # Return only the UUID
            return attachment_id

    @traced(name="jobs_create_attachment", run_type="uipath")
    async def create_attachment_async(
        self,
        *,
        name: str,
        content: Optional[Union[str, bytes]] = None,
        source_path: Optional[Union[str, Path]] = None,
        job_key: Optional[Union[str, uuid.UUID]] = None,
        category: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> uuid.UUID:
        """Create and upload an attachment asynchronously, optionally linking it to a job.

        This method asynchronously handles creating an attachment from a file or memory data.
        If a job key is provided or available in the execution context, the attachment
        will be created in UiPath and linked to the job. If no job is available,
        the file will be saved to a temporary storage folder.

        Note:
            The local storage functionality (when no job is available) is intended for
            local development and debugging purposes only.

        Args:
            name (str): The name of the attachment file.
            content (Optional[Union[str, bytes]]): The content to upload (string or bytes).
            source_path (Optional[Union[str, Path]]): The local path of the file to upload.
            job_key (Optional[Union[str, uuid.UUID]]): The key of the job to link the attachment to.
            category (Optional[str]): Optional category for the attachment in the context of the job.
            folder_key (Optional[str]): The key of the folder. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder. Override the default one set in the SDK config.

        Returns:
            uuid.UUID: The unique identifier for the created attachment, regardless of whether it was
                uploaded to UiPath or stored locally.

        Raises:
            ValueError: If neither content nor source_path is provided, or if both are provided.
            Exception: If the upload fails.

        Examples:
            ```python
            import asyncio
            from uipath.platform import UiPath

            client = UiPath()

            async def main():
                # Create attachment from file and link to job
                attachment_id = await client.jobs.create_attachment_async(
                    name="document.pdf",
                    source_path="path/to/local/document.pdf",
                    job_key="38073051"
                )
                print(f"Created and linked attachment: {attachment_id}")

                # Create attachment from memory content (no job available - saves to temp storage)
                attachment_id = await client.jobs.create_attachment_async(
                    name="report.txt",
                    content="This is a text report"
                )
                print(f"Created attachment: {attachment_id}")
            ```
        """
        # Validate input parameters
        if not (content or source_path):
            raise ValueError("Content or source_path is required")
        if content and source_path:
            raise ValueError("Content and source_path are mutually exclusive")

        # Get job key from context if not explicitly provided
        context_job_key = None
        if job_key is None:
            try:
                context_job_key = self._execution_context.instance_key
            except ValueError:
                # Instance key is not set in environment
                context_job_key = None

        # Check if a job is available
        if job_key is not None or context_job_key is not None:
            # Job is available - create attachment in UiPath and link to job
            actual_job_key = job_key if job_key is not None else context_job_key

            # Create the attachment using the attachments service
            if content is not None:
                attachment_key = await self._attachments_service.upload_async(
                    name=name,
                    content=content,
                    folder_key=folder_key,
                    folder_path=folder_path,
                )
            else:
                # source_path must be provided due to validation check above
                attachment_key = await self._attachments_service.upload_async(
                    name=name,
                    source_path=cast(str, source_path),
                    folder_key=folder_key,
                    folder_path=folder_path,
                )

            # Convert to UUID if string
            if isinstance(actual_job_key, str):
                actual_job_key = uuid.UUID(actual_job_key)

            # Link attachment to job
            await self.link_attachment_async(
                attachment_key=attachment_key,
                job_key=cast(uuid.UUID, actual_job_key),
                category=category,
                folder_key=folder_key,
                folder_path=folder_path,
            )

            return attachment_key
        else:
            # No job available - save to temp folder
            # Generate a UUID to use as identifier
            attachment_id = uuid.uuid4()

            # Create destination file path
            dest_path = os.path.join(self._temp_dir, f"{attachment_id}_{name}")

            # If we have source_path, copy the file
            if source_path is not None:
                source_path_str = (
                    source_path if isinstance(source_path, str) else str(source_path)
                )
                shutil.copy2(source_path_str, dest_path)
            # If we have content, write it to a file
            elif content is not None:
                # Convert string to bytes if needed
                if isinstance(content, str):
                    content = content.encode("utf-8")

                with open(dest_path, "wb") as f:
                    f.write(content)

            # Return only the UUID
            return attachment_id

    def _list_spec(
        self,
        folder_path: Optional[str],
        folder_key: Optional[str],
        filter: Optional[str],
        orderby: Optional[str],
        skip: int,
        top: int,
    ) -> RequestSpec:
        """Build OData request for listing jobs."""
        params: Dict[str, Any] = {"$skip": skip, "$top": top}
        if filter:
            params["$filter"] = filter
        if orderby:
            params["$orderby"] = orderby

        return RequestSpec(
            method="GET",
            endpoint=Endpoint("/orchestrator_/odata/Jobs"),
            params=params,
            headers={**header_folder(folder_key, folder_path)},
        )

    def _stop_spec(
        self,
        job_ids: List[int],
        strategy: str,
        folder_key: Optional[str],
        folder_path: Optional[str],
    ) -> RequestSpec:
        """Build request for stopping jobs with strategy.

        Pure function - no HTTP calls, no side effects.

        Args:
            job_ids: List of job integer IDs (already resolved)
            strategy: Stop strategy ("SoftStop" or "Kill")
            folder_key: Folder key
            folder_path: Folder path

        Returns:
            RequestSpec: Request specification for StopJobs endpoint
        """
        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.StopJobs"
            ),
            json={
                "jobIds": job_ids,
                "strategy": strategy,
            },
            headers={**header_folder(folder_key, folder_path)},
        )

    def _restart_spec(
        self,
        job_id: int,
        folder_key: Optional[str],
        folder_path: Optional[str],
    ) -> RequestSpec:
        """Build request for restarting a job."""
        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.RestartJob"
            ),
            json={"jobId": job_id},
            headers={**header_folder(folder_key, folder_path)},
        )

    def _resolve_job_identifiers(
        self,
        job_keys: List[str],
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> List[int]:
        """Resolve job keys to job IDs in chunked bulk queries.

        This method prevents N+1 query issues by fetching all job IDs using
        OData $filter with 'in' operator. Requests are chunked to avoid URL
        length limits (max 50 keys per request).

        Args:
            job_keys: List of job UUID keys to resolve
            folder_key: Folder key
            folder_path: Folder path

        Returns:
            List[int]: List of job integer IDs in same order as job_keys

        Raises:
            ValueError: If any job key is not a valid UUID
            LookupError: If any job key is not found

        Note:
            Duplicate keys in input are allowed and will return corresponding IDs.
        """
        if not job_keys:
            return []

        for key in job_keys:
            try:
                uuid.UUID(key)
            except ValueError as e:
                raise ValueError(f"Invalid job key format: {key}") from e

        unique_keys = []
        seen = set()
        for key in job_keys:
            if key not in seen:
                unique_keys.append(key)
                seen.add(key)

        CHUNK_SIZE = 50
        all_key_to_id: Dict[str, int] = {}

        for i in range(0, len(unique_keys), CHUNK_SIZE):
            chunk = unique_keys[i : i + CHUNK_SIZE]
            keys_formatted = "','".join(chunk)
            filter_expr = f"Key in ('{keys_formatted}')"

            spec = RequestSpec(
                method="GET",
                endpoint=Endpoint("/orchestrator_/odata/Jobs"),
                params={
                    "$filter": filter_expr,
                    "$select": "Id,Key",
                    "$top": len(chunk),
                },
                headers={**header_folder(folder_key, folder_path)},
            )

            response = self.request(
                spec.method,
                url=spec.endpoint,
                params=spec.params,
                headers=spec.headers,
            ).json()

            items = response.get("value", [])

            # Accumulate mappings from this chunk
            for item in items:
                all_key_to_id[item["Key"]] = item["Id"]

        # Verify all unique keys were found
        if len(all_key_to_id) != len(unique_keys):
            found_keys = set(all_key_to_id.keys())
            missing_keys = set(unique_keys) - found_keys
            raise LookupError(f"Jobs not found for keys: {', '.join(missing_keys)}")

        # Build result preserving original order (including duplicates)
        return [all_key_to_id[key] for key in job_keys]

    async def _resolve_job_identifiers_async(
        self,
        job_keys: List[str],
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> List[int]:
        """Async version of _resolve_job_identifiers()."""
        if not job_keys:
            return []

        for key in job_keys:
            try:
                uuid.UUID(key)
            except ValueError as e:
                raise ValueError(f"Invalid job key format: {key}") from e

        unique_keys = []
        seen = set()
        for key in job_keys:
            if key not in seen:
                unique_keys.append(key)
                seen.add(key)

        CHUNK_SIZE = 50
        all_key_to_id: Dict[str, int] = {}

        for i in range(0, len(unique_keys), CHUNK_SIZE):
            chunk = unique_keys[i : i + CHUNK_SIZE]
            keys_formatted = "','".join(chunk)
            filter_expr = f"Key in ('{keys_formatted}')"

            spec = RequestSpec(
                method="GET",
                endpoint=Endpoint("/orchestrator_/odata/Jobs"),
                params={
                    "$filter": filter_expr,
                    "$select": "Id,Key",
                    "$top": len(chunk),
                },
                headers={**header_folder(folder_key, folder_path)},
            )

            response = (
                await self.request_async(
                    spec.method,
                    url=spec.endpoint,
                    params=spec.params,
                    headers=spec.headers,
                )
            ).json()

            items = response.get("value", [])

            for item in items:
                all_key_to_id[item["Key"]] = item["Id"]

        if len(all_key_to_id) != len(unique_keys):
            found_keys = set(all_key_to_id.keys())
            missing_keys = set(unique_keys) - found_keys
            raise LookupError(f"Jobs not found for keys: {', '.join(missing_keys)}")

        return [all_key_to_id[key] for key in job_keys]



from dataclasses import dataclass
from functools import cached_property

from httpx import HTTPStatusError


@dataclass(frozen=True)
class ExtractedErrorInfo:
    message: str | None = None
    error_code: str | None = None
    trace_id: str | None = None


class EnrichedException(Exception):
    """Enriched HTTP error with detailed request/response information.

    Wraps HTTPStatusError with URL, method, status code, and truncated response
    content in __str__. For structured error fields, use ``error_info`` which
    delegates to per-service extractors.

    Does not retain a reference to the original HTTPStatusError — all needed
    data is eagerly extracted. Callers needing the raw response can still
    access ``__cause__`` (set by ``raise EnrichedException(e) from e``).
    """

    def __init__(self, error: HTTPStatusError) -> None:
        # while status code 0 is the correct one according to http standard;
        # it has a totally oposite meaning as return codes in CLIs;
        # opted for -1 to avoid confusion
        self.status_code: int = error.response.status_code if error.response else -1
        self.url: str = str(error.request.url) if error.request else "Unknown"
        self.http_method: str = (
            error.request.method
            if error.request and error.request.method
            else "Unknown"
        )

        self._response_body: str | None = None
        self._content_type: str | None = None
        if error.response is not None:
            self._content_type = error.response.headers.get("content-type")
            if error.response.content:
                try:
                    self._response_body = error.response.content.decode("utf-8")
                except Exception:
                    pass

        max_content_length = 200
        if self._response_body:
            if len(self._response_body) > max_content_length:
                self.response_content = (
                    self._response_body[:max_content_length] + "... (truncated)"
                )
            else:
                self.response_content = self._response_body
        else:
            self.response_content = "No content"

        enriched_message = (
            f"\nRequest URL: {self.url}"
            f"\nHTTP Method: {self.http_method}"
            f"\nStatus Code: {self.status_code}"
            f"\nResponse Content: {self.response_content}"
        )

        super().__init__(enriched_message)

    @cached_property
    def error_info(self) -> ExtractedErrorInfo | None:
        """Service-aware extraction of message, error_code, trace_id."""
        from ._extractors._router import extract_error_info

        if not self._response_body:
            return None
        return extract_error_info(self.url, self._response_body, self._content_type)



class FolderNotFoundException(Exception):
    """Raised when a folder cannot be found.

    This exception is raised when attempting to access a folder that does not exist
    in the UiPath Orchestrator.
    """

    def __init__(
        self,
        folder_name,
    ):
        self.message = f"Folder {folder_name} not found."
        super().__init__(self.message)



from unittest.mock import PropertyMock, patch

import pytest

from uipath.platform.common._config import ConfigurationManager
from uipath.platform.common._http_config import get_httpx_client_kwargs


@pytest.fixture(autouse=True)
def _clean_env(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensure licensing-related env vars are clean for every test."""
    monkeypatch.delenv("UIPATH_DISABLE_SSL_VERIFY", raising=False)


class TestGetHttpxClientKwargsHeaders:
    """Tests for header merging in get_httpx_client_kwargs()."""

    def test_no_licensing_no_caller_headers(self) -> None:
        """No headers key when neither licensing nor caller headers are provided."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value=None,
        ):
            result = get_httpx_client_kwargs()
        assert "headers" not in result

    def test_licensing_context_only(self) -> None:
        """Licensing header included when licensing_context is set."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value="test-license-ctx",
        ):
            result = get_httpx_client_kwargs()
        assert result["headers"] == {"x-uipath-licensing-context": "test-license-ctx"}

    def test_caller_headers_only(self) -> None:
        """Caller headers included when no licensing_context is set."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value=None,
        ):
            result = get_httpx_client_kwargs(headers={"Authorization": "Bearer tok"})
        assert result["headers"] == {"Authorization": "Bearer tok"}

    def test_licensing_and_caller_headers_merged(self) -> None:
        """Both licensing and caller headers present in result."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value="lic-ctx",
        ):
            result = get_httpx_client_kwargs(headers={"Authorization": "Bearer tok"})
        assert result["headers"] == {
            "x-uipath-licensing-context": "lic-ctx",
            "Authorization": "Bearer tok",
        }

    def test_caller_headers_win_on_conflict(self) -> None:
        """Caller headers override licensing headers on same key."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value="lic-ctx",
        ):
            result = get_httpx_client_kwargs(
                headers={"x-uipath-licensing-context": "caller-override"}
            )
        assert result["headers"] == {"x-uipath-licensing-context": "caller-override"}

    def test_result_always_contains_base_config(self) -> None:
        """SSL, timeout, and redirect config always present regardless of headers."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value="lic",
        ):
            result = get_httpx_client_kwargs(headers={"Authorization": "Bearer x"})
        assert result["follow_redirects"] is True
        assert result["timeout"] == 30.0
        assert "verify" in result

    def test_no_headers_key_when_empty(self) -> None:
        """Empty caller headers dict with no licensing does not add headers key."""
        with patch.object(
            ConfigurationManager,
            "licensing_context",
            new_callable=PropertyMock,
            return_value=None,
        ):
            result = get_httpx_client_kwargs(headers={})
        assert "headers" not in result



import os
from functools import cached_property
from pathlib import Path

from pydantic import BaseModel


class UiPathApiConfig(BaseModel):
    base_url: str
    secret: str


class ConfigurationManager:
    _instance = None
    studio_solution_id: str | None = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @property
    def bindings_file_path(self) -> Path:
        from uipath.platform.common.constants import UIPATH_BINDINGS_FILE

        return Path(UIPATH_BINDINGS_FILE)

    @property
    def config_file_path(self) -> Path:
        from uipath.platform.common.constants import (
            ENV_UIPATH_CONFIG_PATH,
            UIPATH_CONFIG_FILE,
        )

        return Path(os.environ.get(ENV_UIPATH_CONFIG_PATH, UIPATH_CONFIG_FILE))

    @property
    def config_file_name(self) -> str:
        from uipath.platform.common.constants import UIPATH_CONFIG_FILE

        return UIPATH_CONFIG_FILE

    @property
    def project_id(self) -> str | None:
        from uipath.platform.common.constants import ENV_UIPATH_PROJECT_ID

        return os.getenv(ENV_UIPATH_PROJECT_ID, None)

    @property
    def project_key(self) -> str | None:
        from uipath.platform.common.constants import ENV_PROJECT_KEY

        return os.getenv(ENV_PROJECT_KEY, None)

    @property
    def tenant_name(self) -> str | None:
        from uipath.platform.common.constants import ENV_TENANT_NAME

        return os.getenv(ENV_TENANT_NAME, None)

    @property
    def tenant_id(self) -> str | None:
        from uipath.platform.common.constants import ENV_TENANT_ID

        return os.getenv(ENV_TENANT_ID, None)

    @property
    def organization_id(self) -> str | None:
        from uipath.platform.common.constants import ENV_ORGANIZATION_ID

        return os.getenv(ENV_ORGANIZATION_ID, None)

    @property
    def base_url(self) -> str | None:
        from uipath.platform.common.constants import ENV_BASE_URL

        return os.getenv(ENV_BASE_URL, None)

    @property
    def folder_key(self) -> str | None:
        from uipath.platform.common.constants import ENV_FOLDER_KEY

        return os.getenv(ENV_FOLDER_KEY, None)

    @property
    def process_uuid(self) -> str | None:
        from uipath.platform.common.constants import ENV_UIPATH_PROCESS_UUID

        return os.getenv(ENV_UIPATH_PROCESS_UUID, None)

    @property
    def trace_id(self) -> str | None:
        from uipath.platform.common.constants import ENV_UIPATH_TRACE_ID

        return os.getenv(ENV_UIPATH_TRACE_ID, None)

    @property
    def process_version(self) -> str | None:
        from uipath.platform.common.constants import ENV_UIPATH_PROCESS_VERSION

        return os.getenv(ENV_UIPATH_PROCESS_VERSION, None)

    @property
    def is_studio_project(self) -> bool:
        return self.project_id is not None

    @property
    def job_key(self) -> str | None:
        from uipath.platform.common.constants import ENV_JOB_KEY

        return os.getenv(ENV_JOB_KEY, None)

    @property
    def has_legacy_eval_folder(self) -> bool:
        from uipath.platform.common.constants import LEGACY_EVAL_FOLDER

        eval_path = Path(os.getcwd()) / LEGACY_EVAL_FOLDER
        return eval_path.exists() and eval_path.is_dir()

    @property
    def has_eval_folder(self) -> bool:
        from uipath.platform.common.constants import EVALS_FOLDER

        coded_eval_path = Path(os.getcwd()) / EVALS_FOLDER
        return coded_eval_path.exists() and coded_eval_path.is_dir()

    @property
    def entry_points_file_path(self) -> Path:
        from uipath.platform.common.constants import ENTRY_POINTS_FILE

        return Path(ENTRY_POINTS_FILE)

    @property
    def uiproj_file_path(self) -> Path:
        from uipath.platform.common.constants import UIPROJ_FILE

        return Path(UIPROJ_FILE)

    @property
    def studio_metadata_file_path(self) -> Path:
        from uipath.platform.common.constants import STUDIO_METADATA_FILE

        return Path(".uipath", STUDIO_METADATA_FILE)

    @property
    def licensing_context(self) -> str | None:
        return self._read_internal_argument("licensingContext")

    @property
    def is_rooted_to_debug_job(self) -> bool:
        """Whether this job, which may be a deployed process, is rooted to a debug session (e.g. Maestro solution debug)."""
        return self._read_internal_argument("isDebug") is True

    @property
    def is_tracing_enabled(self) -> bool:
        from uipath.platform.common.constants import ENV_TRACING_ENABLED

        return os.getenv(ENV_TRACING_ENABLED, "true").lower() == "true"

    def _read_internal_argument(self, key: str) -> str | None:
        internal_args = self._internal_arguments
        return internal_args.get(key) if internal_args else None

    @cached_property
    def _internal_arguments(self) -> dict[str, str] | None:
        import json

        try:
            with open(self.config_file_path, "r") as f:
                data = json.load(f)
                return data.get("runtime", {}).get("internalArguments")
        except (FileNotFoundError, json.JSONDecodeError):
            return None


UiPathConfig = ConfigurationManager()



from typing import Any

import pytest
from pytest_httpx import HTTPXMock

from uipath.platform import UiPathApiConfig, UiPathExecutionContext
from uipath.platform.action_center import Task
from uipath.platform.action_center._tasks_service import TasksService
from uipath.platform.common.constants import HEADER_USER_AGENT


@pytest.fixture
def service(
    config: UiPathApiConfig,
    execution_context: UiPathExecutionContext,
    monkeypatch: pytest.MonkeyPatch,
) -> TasksService:
    monkeypatch.setenv("UIPATH_FOLDER_PATH", "test-folder-path")

    return TasksService(config=config, execution_context=execution_context)


class TestTasksService:
    def test_retrieve(
        self,
        httpx_mock: HTTPXMock,
        service: TasksService,
        base_url: str,
        org: str,
        tenant: str,
        version: str,
    ) -> None:
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/tasks/GenericTasks/GetTaskDataByKey?taskKey=test-id",
            status_code=200,
            json={"id": 1, "title": "Test Action"},
        )

        action = service.retrieve(
            action_key="test-id",
            app_folder_path="test-folder",
        )

        assert isinstance(action, Task)
        assert action.id == 1
        assert action.title == "Test Action"

        sent_request = httpx_mock.get_request()
        if sent_request is None:
            raise Exception("No request was sent")

        assert sent_request.method == "GET"
        assert (
            sent_request.url
            == f"{base_url}{org}{tenant}/orchestrator_/tasks/GenericTasks/GetTaskDataByKey?taskKey=test-id"
        )

        assert HEADER_USER_AGENT in sent_request.headers
        assert (
            sent_request.headers[HEADER_USER_AGENT]
            == f"UiPath.Python.Sdk/UiPath.Python.Sdk.Activities.TasksService.retrieve/{version}"
        )

    @pytest.mark.anyio
    async def test_retrieve_async(
        self,
        httpx_mock: HTTPXMock,
        service: TasksService,
        base_url: str,
        org: str,
        tenant: str,
        version: str,
    ) -> None:
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/tasks/GenericTasks/GetTaskDataByKey?taskKey=test-id",
            status_code=200,
            json={"id": 1, "title": "Test Action"},
        )

        action = await service.retrieve_async(
            action_key="test-id",
            app_folder_path="test-folder",
        )

        assert isinstance(action, Task)
        assert action.id == 1
        assert action.title == "Test Action"

        sent_request = httpx_mock.get_request()
        if sent_request is None:
            raise Exception("No request was sent")

        assert sent_request.method == "GET"
        assert (
            sent_request.url
            == f"{base_url}{org}{tenant}/orchestrator_/tasks/GenericTasks/GetTaskDataByKey?taskKey=test-id"
        )

        assert HEADER_USER_AGENT in sent_request.headers
        assert (
            sent_request.headers[HEADER_USER_AGENT]
            == f"UiPath.Python.Sdk/UiPath.Python.Sdk.Activities.TasksService.retrieve_async/{version}"
        )

    def test_create_with_app_key(
        self,
        httpx_mock: HTTPXMock,
        service: TasksService,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/tasks/AppTasks/CreateAppTask",
            status_code=200,
            json={"id": 1, "title": "Test Action"},
        )

        action = service.create(
            title="Test Action",
            app_key="test-app-key",
            data={"test": "data"},
        )

        assert isinstance(action, Task)
        assert action.id == 1
        assert action.title == "Test Action"

    def test_create_with_assignee(
        self,
        httpx_mock: HTTPXMock,
        service: TasksService,
        base_url: str,
        org: str,
        tenant: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")

        httpx_mock.add_response(
            url=f"{base_url}{org}/apps_/default/api/v1/default/deployed-action-apps-schemas?search=test-app&filterByDeploymentTitle=true",
            status_code=200,
            json={
                "deployed": [
                    {
                        "systemName": "test-app",
                        "deploymentTitle": "test-app",
                        "actionSchema": {
                            "key": "test-key",
                            "inputs": [],
                            "outputs": [],
                            "inOuts": [],
                            "outcomes": [],
                        },
                        "deploymentFolder": {
                            "fullyQualifiedName": "test-folder-path",
                            "key": "test-folder-key",
                        },
                    }
                ]
            },
        )

        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/tasks/AppTasks/CreateAppTask",
            status_code=200,
            json={"id": 1, "title": "Test Action"},
        )

        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Tasks/UiPath.Server.Configuration.OData.AssignTasks",
            status_code=200,
            json={},
        )

        action = service.create(
            title="Test Action",
            app_name="test-app",
            data={"test": "data"},
            assignee="test@example.com",
        )

        assert isinstance(action, Task)
        assert action.id == 1
        assert action.title == "Test Action"


def _make_deployed_app(
    name: str,
    folder_path: str,
    folder_key: str,
    system_name: str | None = None,
) -> dict[str, Any]:
    return {
        "systemName": system_name or name,
        "deploymentTitle": name,
        "actionSchema": {
            "key": f"{name}-schema-key",
            "inputs": [],
            "outputs": [],
            "inOuts": [],
            "outcomes": [],
        },
        "deploymentFolder": {
            "fullyQualifiedName": folder_path,
            "key": folder_key,
        },
    }


class TestCreateFiltersByFolder:
    """Tests for folder filtering when creating tasks via app_name lookup."""

    def _make_tasks_service(
        self,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        folder_key: str | None = None,
        folder_path: str | None = None,
    ) -> TasksService:
        if folder_key:
            monkeypatch.setenv("UIPATH_FOLDER_KEY", folder_key)
        if folder_path:
            monkeypatch.setenv("UIPATH_FOLDER_PATH", folder_path)
        return TasksService(config=config, execution_context=execution_context)

    def _mock_app_schemas_response(
        self,
        httpx_mock: HTTPXMock,
        base_url: str,
        org: str,
        app_name: str,
        deployed_apps: list[dict[str, Any]],
    ) -> None:
        httpx_mock.add_response(
            url=f"{base_url}{org}/apps_/default/api/v1/default/deployed-action-apps-schemas?search={app_name}&filterByDeploymentTitle=true",
            status_code=200,
            json={"deployed": deployed_apps},
        )

    def _mock_create_task_response(
        self,
        httpx_mock: HTTPXMock,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/tasks/AppTasks/CreateAppTask",
            status_code=200,
            json={"id": 1, "title": "Test Action"},
        )

    def test_create_filters_by_app_folder_key(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(config, execution_context, monkeypatch)
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [
                _make_deployed_app(
                    "my-app", "folder-a", "key-a", system_name="my-app-a"
                ),
                _make_deployed_app(
                    "my-app", "folder-b", "key-b", system_name="my-app-b"
                ),
            ],
        )
        self._mock_create_task_response(httpx_mock, base_url, org, tenant)

        task = tasks_service.create(
            title="Test",
            app_name="my-app",
            app_folder_key="key-b",
            app_folder_path=None,
        )

        assert isinstance(task, Task)
        create_request = [
            r for r in httpx_mock.get_requests() if "CreateAppTask" in str(r.url)
        ][0]
        # systemName from the key-b deployment is used as appId in the request
        assert b"my-app-b" in create_request.content

    @pytest.mark.anyio
    async def test_create_async_filters_by_app_folder_key(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(config, execution_context, monkeypatch)
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [
                _make_deployed_app(
                    "my-app", "folder-a", "key-a", system_name="my-app-a"
                ),
                _make_deployed_app(
                    "my-app", "folder-b", "key-b", system_name="my-app-b"
                ),
            ],
        )
        self._mock_create_task_response(httpx_mock, base_url, org, tenant)

        task = await tasks_service.create_async(
            title="Test",
            app_name="my-app",
            app_folder_key="key-b",
            app_folder_path=None,
        )

        assert isinstance(task, Task)
        create_request = [
            r for r in httpx_mock.get_requests() if "CreateAppTask" in str(r.url)
        ][0]
        # systemName from the key-b deployment is used as appId in the request
        assert b"my-app-b" in create_request.content

    def test_create_filters_by_app_folder_path(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(config, execution_context, monkeypatch)
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [
                _make_deployed_app("my-app", "folder-a", "key-a"),
                _make_deployed_app("my-app", "folder-b", "key-b"),
            ],
        )
        self._mock_create_task_response(httpx_mock, base_url, org, tenant)

        task = tasks_service.create(
            title="Test",
            app_name="my-app",
            app_folder_path="folder-a",
        )

        assert isinstance(task, Task)

    def test_create_falls_back_to_env_folder_path(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(
            config, execution_context, monkeypatch, folder_path="env-folder-path"
        )
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [
                _make_deployed_app("my-app", "folder-a", "key-a"),
                _make_deployed_app(
                    "my-app", "env-folder-path", "key-b", system_name="my-app-b"
                ),
            ],
        )
        self._mock_create_task_response(httpx_mock, base_url, org, tenant)

        task = tasks_service.create(
            title="Test",
            app_name="my-app",
        )

        assert isinstance(task, Task)
        create_request = [
            r for r in httpx_mock.get_requests() if "CreateAppTask" in str(r.url)
        ][0]
        assert b"my-app-b" in create_request.content

    def test_create_falls_back_to_env_folder_key(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(
            config, execution_context, monkeypatch, folder_key="env-key"
        )
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [
                _make_deployed_app("my-app", "folder-a", "key-a"),
                _make_deployed_app("my-app", "folder-b", "env-key"),
            ],
        )
        self._mock_create_task_response(httpx_mock, base_url, org, tenant)

        task = tasks_service.create(
            title="Test",
            app_name="my-app",
        )

        assert isinstance(task, Task)

    def test_create_raises_when_app_not_found_in_tenant(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(
            config, execution_context, monkeypatch, folder_path="folder-a"
        )
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [_make_deployed_app("other-app", "folder-a", "key-a")],
        )

        with pytest.raises(
            Exception, match="'my-app' was not found in the current tenant"
        ):
            tasks_service.create(title="Test", app_name="my-app")

    def test_create_raises_when_app_not_found_in_folder_path(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(config, execution_context, monkeypatch)
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [_make_deployed_app("my-app", "folder-a", "key-a")],
        )

        with pytest.raises(
            Exception,
            match="'my-app' was not found in folder with fully qualified name 'folder-b'",
        ):
            tasks_service.create(
                title="Test", app_name="my-app", app_folder_path="folder-b"
            )

    def test_create_raises_when_app_not_found_in_folder_key(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(config, execution_context, monkeypatch)
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [_make_deployed_app("my-app", "folder-a", "key-a")],
        )

        with pytest.raises(
            Exception,
            match="'my-app' was not found in folder with identifier 'wrong-key'",
        ):
            tasks_service.create(
                title="Test",
                app_name="my-app",
                app_folder_key="wrong-key",
                app_folder_path=None,
            )

    def test_create_raises_when_no_folder_key_or_path_provided(
        self,
        httpx_mock: HTTPXMock,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        monkeypatch: pytest.MonkeyPatch,
        base_url: str,
        org: str,
        tenant: str,
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "test-tenant-id")
        tasks_service = self._make_tasks_service(config, execution_context, monkeypatch)
        self._mock_app_schemas_response(
            httpx_mock,
            base_url,
            org,
            "my-app",
            [_make_deployed_app("my-app", "folder-a", "key-a")],
        )

        with pytest.raises(
            Exception, match="no folder key or folder path was provided"
        ):
            tasks_service.create(
                title="Test",
                app_name="my-app",
                app_folder_path=None,
            )



from unittest.mock import AsyncMock, patch

import httpx
import pytest

from uipath.platform.common._external_application_service import (
    ExternalApplicationService,
)
from uipath.platform.common.auth import TokenData
from uipath.platform.errors import EnrichedException

IDENTITY_SERVICE_SYNC = (
    "uipath.platform.identity.IdentityService.get_client_credentials_token"
)
IDENTITY_SERVICE_ASYNC = (
    "uipath.platform.identity.IdentityService.get_client_credentials_token_async"
)


class TestExternalApplicationService:
    @pytest.mark.parametrize(
        "url,expected_domain",
        [
            ("https://alpha.uipath.com", "alpha"),
            ("https://sub.alpha.uipath.com", "alpha"),
            ("https://staging.uipath.com", "staging"),
            ("https://env.staging.uipath.com", "staging"),
            ("https://cloud.uipath.com", "cloud"),
            ("https://org.cloud.uipath.com", "cloud"),
            ("https://something-else.com", "cloud"),
            ("not-a-url", "cloud"),
        ],
    )
    def test_extract_domain_from_base_url(self, url: str, expected_domain: str):
        service = ExternalApplicationService(url)
        assert service._domain == expected_domain

    @pytest.mark.parametrize("use_async", [False, True])
    async def test_get_access_token_success(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        fake_token = TokenData(access_token="fake-token")

        if use_async:
            with patch(
                IDENTITY_SERVICE_ASYNC, new_callable=AsyncMock, return_value=fake_token
            ):
                token = await service.get_token_data_async("client-id", "client-secret")
        else:
            with patch(IDENTITY_SERVICE_SYNC, return_value=fake_token):
                token = service.get_token_data("client-id", "client-secret")

        assert token.access_token == "fake-token"

    @pytest.mark.parametrize("use_async", [False, True])
    async def test_get_access_token_invalid_client(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        response = httpx.Response(
            400, json={}, request=httpx.Request("POST", "http://test")
        )
        side_effect = httpx.HTTPStatusError(
            "", request=response.request, response=response
        )

        with pytest.raises(EnrichedException) as exc:
            if use_async:
                with patch(
                    IDENTITY_SERVICE_ASYNC,
                    new_callable=AsyncMock,
                    side_effect=side_effect,
                ):
                    await service.get_token_data_async("bad-id", "bad-secret")
            else:
                with patch(IDENTITY_SERVICE_SYNC, side_effect=side_effect):
                    service.get_token_data("bad-id", "bad-secret")

        assert "400" in str(exc.value)

    @pytest.mark.parametrize("use_async", [False, True])
    async def test_get_access_token_unauthorized(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        response = httpx.Response(
            401, json={}, request=httpx.Request("POST", "http://test")
        )
        side_effect = httpx.HTTPStatusError(
            "", request=response.request, response=response
        )

        with pytest.raises(EnrichedException) as exc:
            if use_async:
                with patch(
                    IDENTITY_SERVICE_ASYNC,
                    new_callable=AsyncMock,
                    side_effect=side_effect,
                ):
                    await service.get_token_data_async("bad-id", "bad-secret")
            else:
                with patch(IDENTITY_SERVICE_SYNC, side_effect=side_effect):
                    service.get_token_data("bad-id", "bad-secret")

        assert "401" in str(exc.value)

    @pytest.mark.parametrize("use_async", [False, True])
    async def test_get_access_token_unexpected_status(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        response = httpx.Response(
            500, json={}, request=httpx.Request("POST", "http://test")
        )
        side_effect = httpx.HTTPStatusError(
            "", request=response.request, response=response
        )

        with pytest.raises(EnrichedException) as exc:
            if use_async:
                with patch(
                    IDENTITY_SERVICE_ASYNC,
                    new_callable=AsyncMock,
                    side_effect=side_effect,
                ):
                    await service.get_token_data_async("client-id", "client-secret")
            else:
                with patch(IDENTITY_SERVICE_SYNC, side_effect=side_effect):
                    service.get_token_data("client-id", "client-secret")

        assert "500" in str(exc.value)

    @pytest.mark.parametrize("use_async", [False, True])
    async def test_get_access_token_network_error(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        side_effect = httpx.RequestError("network down")

        with pytest.raises(Exception) as exc:
            if use_async:
                with patch(
                    IDENTITY_SERVICE_ASYNC,
                    new_callable=AsyncMock,
                    side_effect=side_effect,
                ):
                    await service.get_token_data_async("client-id", "client-secret")
            else:
                with patch(IDENTITY_SERVICE_SYNC, side_effect=side_effect):
                    service.get_token_data("client-id", "client-secret")

        assert "Network error during authentication" in str(exc.value)

    @pytest.mark.parametrize("use_async", [False, True])
    async def test_get_access_token_unexpected_exception(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        side_effect = ValueError("weird error")

        with pytest.raises(Exception) as exc:
            if use_async:
                with patch(
                    IDENTITY_SERVICE_ASYNC,
                    new_callable=AsyncMock,
                    side_effect=side_effect,
                ):
                    await service.get_token_data_async("client-id", "client-secret")
            else:
                with patch(IDENTITY_SERVICE_SYNC, side_effect=side_effect):
                    service.get_token_data("client-id", "client-secret")

        assert "Unexpected error during authentication" in str(exc.value)


class TestExternalApplicationServiceDelegation:
    @pytest.mark.parametrize("use_async", [False, True])
    async def test_delegates_to_identity_service(self, use_async: bool):
        service = ExternalApplicationService("https://cloud.uipath.com")
        fake_token = TokenData(access_token="delegated-token")

        if use_async:
            with patch.object(
                service._identity_service,
                "get_client_credentials_token_async",
                new_callable=AsyncMock,
                return_value=fake_token,
            ) as mock_get_token:
                result = await service.get_token_data_async(
                    "my-id", "my-secret", scope="OR.Execution"
                )

            mock_get_token.assert_called_once_with(
                client_id="my-id",
                client_secret="my-secret",
                scope="OR.Execution",
            )
        else:
            with patch.object(
                service._identity_service,
                "get_client_credentials_token",
                return_value=fake_token,
            ) as mock_get_token:
                result = service.get_token_data(
                    "my-id", "my-secret", scope="OR.Execution"
                )

            mock_get_token.assert_called_once_with(
                client_id="my-id",
                client_secret="my-secret",
                scope="OR.Execution",
            )
        assert result.access_token == "delegated-token"

    @pytest.mark.parametrize(
        "base_url,expected_domain",
        [
            ("https://alpha.uipath.com", "https://alpha.uipath.com"),
            ("https://sub.alpha.uipath.com", "https://alpha.uipath.com"),
            ("https://staging.uipath.com", "https://staging.uipath.com"),
            ("https://env.staging.uipath.com", "https://staging.uipath.com"),
            ("https://cloud.uipath.com", "https://cloud.uipath.com"),
            ("https://org.cloud.uipath.com", "https://cloud.uipath.com"),
            ("https://something-else.com", "https://cloud.uipath.com"),
        ],
    )
    def test_domain_mapping_for_delegation(self, base_url: str, expected_domain: str):
        service = ExternalApplicationService(base_url)
        assert service._identity_service._domain == expected_domain



"""Automation Tracker (BTS) models for UiPath Platform.

Models for tracking business transactions and operations
via the Business Transaction Service, used for Process Mining.
"""

from datetime import datetime
from enum import Enum
from typing import Dict, Optional

from pydantic import BaseModel, ConfigDict, Field


class TransactionStatus(str, Enum):
    """Status of a BTS transaction."""

    UNKNOWN = "Unknown"
    SUCCESSFUL = "Successful"
    FAILED = "Failed"


class OperationStatus(str, Enum):
    """Status of a BTS operation."""

    UNKNOWN = "Unknown"
    SUCCESSFUL = "Successful"
    FAILED = "Failed"
    CANCELLED = "Cancelled"
    ABANDONED = "Abandoned"


class TransactionPayload(BaseModel):
    """Wire-format payload for BTS transaction start/end endpoints."""

    model_config = ConfigDict(populate_by_name=True)

    organization_id: str = Field(alias="organizationId")
    tenant_id: str = Field(alias="tenantId")
    transaction_id: str = Field(alias="transactionId")
    name: str
    reference: str
    timestamp: datetime
    fingerprint: str
    result: Optional[str] = None
    status: str
    attributes: Dict[str, str] = Field(default_factory=dict)


class OperationPayload(BaseModel):
    """Wire-format payload for BTS operation start/end endpoints."""

    model_config = ConfigDict(populate_by_name=True)

    organization_id: str = Field(alias="organizationId")
    tenant_id: str = Field(alias="tenantId")
    transaction_id: str = Field(alias="transactionId")
    operation_id: str = Field(alias="operationId")
    parent_operation: Optional[str] = Field(default=None, alias="parentOperation")
    name: str
    timestamp: datetime
    status: str
    attributes: Dict[str, str] = Field(default_factory=dict)
    result: Optional[str] = None
    fingerprint: str



"""Shared helpers for error payload extractors."""

from typing import Any, TypeVar
from urllib.parse import urlparse

T = TypeVar("T")


def get_field(body: dict[str, Any], *keys: str) -> Any:
    """Return the first non-None value found for the given keys.

    Automatically tries both camelCase and PascalCase for each key.
    """
    for key in keys:
        val = body.get(key)
        if val is not None:
            return val
        val = body.get(key[0].swapcase() + key[1:])
        if val is not None:
            return val
    return None


def get_typed_field(body: dict[str, Any], type_: type[T], *keys: str) -> T | None:
    """Return the first non-None value matching *type_* for the given keys.

    Skips values that don't match the expected type, trying the next key.
    """
    for key in keys:
        val = body.get(key)
        if val is None:
            val = body.get(key[0].swapcase() + key[1:])
        if isinstance(val, type_):
            return val
    return None


def get_str_field(body: dict[str, Any], *keys: str) -> str | None:
    """Return the first scalar value for the given keys, converted to str.

    Skips dicts, lists, and None.
    """
    for key in keys:
        val = body.get(key)
        if val is None:
            val = body.get(key[0].swapcase() + key[1:])
        if val is not None and not isinstance(val, (dict, list)):
            return str(val)
    return None


def extract_service_prefix(url: str) -> str | None:
    """Extract the service prefix (e.g. 'orchestrator_') from a UiPath URL path."""
    path = urlparse(url).path if "://" in url else url
    for segment in path.strip("/").split("/"):
        if segment.endswith("_") and len(segment) > 1:
            return segment
    return None


def is_llm_path(url: str) -> bool:
    """Check if the URL targets the LLM Gateway."""
    return "/llm/" in url



import pytest

from uipath.platform.common._service_url_overrides import (
    inject_routing_headers,
    resolve_service_url,
)


class TestResolveServiceUrl:
    def test_returns_none_when_no_override(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.delenv("UIPATH_SERVICE_URL_AGENTHUB", raising=False)
        assert resolve_service_url("agenthub_/llm/api/chat/completions") is None

    def test_agenthub_override(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_AGENTHUB", "http://localhost:5200")
        result = resolve_service_url("agenthub_/llm/api/chat/completions")
        assert result == "http://localhost:5200/llm/api/chat/completions"

    def test_orchestrator_override(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_ORCHESTRATOR", "http://localhost:8080")
        result = resolve_service_url("orchestrator_/odata/Buckets")
        assert result == "http://localhost:8080/odata/Buckets"

    def test_leading_slash_stripped(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_AGENTHUB", "http://localhost:5200")
        result = resolve_service_url("/agenthub_/llm/api/chat/completions")
        assert result == "http://localhost:5200/llm/api/chat/completions"

    def test_trailing_slash_on_override_url(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_AGENTHUB", "http://localhost:5200/")
        result = resolve_service_url("agenthub_/llm/api/chat/completions")
        assert result == "http://localhost:5200/llm/api/chat/completions"

    def test_compound_service_prefix(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_AGENTSRUNTIME", "http://localhost:6100")
        result = resolve_service_url("agentsruntime_/api/execution/guardrails/validate")
        assert result == "http://localhost:6100/api/execution/guardrails/validate"

    def test_no_service_prefix_returns_none(self) -> None:
        assert resolve_service_url("api/v1/users") is None

    def test_empty_path_returns_none(self) -> None:
        assert resolve_service_url("") is None

    def test_service_only_path(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_AGENTHUB", "http://localhost:5200")
        result = resolve_service_url("agenthub_/")
        assert result == "http://localhost:5200/"

    def test_query_params_preserved(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_SERVICE_URL_AGENTHUB", "http://localhost:5200")
        result = resolve_service_url(
            "agenthub_/llm/openai/deployments/gpt-4/chat/completions?api-version=2024-10-21"
        )
        assert (
            result
            == "http://localhost:5200/llm/openai/deployments/gpt-4/chat/completions?api-version=2024-10-21"
        )


class TestInjectRoutingHeaders:
    def test_injects_tenant_and_org(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "tenant-123")
        monkeypatch.setenv("UIPATH_ORGANIZATION_ID", "org-456")
        headers: dict[str, str] = {}
        inject_routing_headers(headers)
        assert headers["X-UiPath-Internal-TenantId"] == "tenant-123"
        assert headers["X-UiPath-Internal-AccountId"] == "org-456"

    def test_skips_missing_env_vars(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.delenv("UIPATH_TENANT_ID", raising=False)
        monkeypatch.delenv("UIPATH_ORGANIZATION_ID", raising=False)
        headers: dict[str, str] = {}
        inject_routing_headers(headers)
        assert "X-UiPath-Internal-TenantId" not in headers
        assert "X-UiPath-Internal-AccountId" not in headers

    def test_does_not_overwrite_existing_headers(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("UIPATH_TENANT_ID", "tenant-123")
        monkeypatch.setenv("UIPATH_ORGANIZATION_ID", "org-456")
        headers: dict[str, str] = {"X-Custom": "keep-me"}
        inject_routing_headers(headers)
        assert headers["X-Custom"] == "keep-me"
        assert headers["X-UiPath-Internal-TenantId"] == "tenant-123"



"""RFC 7807 Problem Details error payload extractor."""

from typing import Any

from .._enriched_exception import ExtractedErrorInfo
from ._helpers import get_typed_field

_RFC7807_CONTENT_TYPES = frozenset(
    {
        "application/problem+json",
        "application/problem+xml",
    }
)


_RFC7807_REQUIRED = {"detail", "title"}
_RFC7807_SUPPORTING = {"type", "status", "instance"}


def is_rfc7807_content_type(content_type: str | None) -> bool:
    """Check if the content type indicates an RFC 7807 response."""
    if not content_type:
        return False
    return content_type.split(";")[0].strip().lower() in _RFC7807_CONTENT_TYPES


def looks_like_rfc7807(body: dict[str, Any]) -> bool:
    """Structural heuristic: has detail/title + at least 2 RFC 7807 keys."""
    keys = body.keys()
    has_required = bool(_RFC7807_REQUIRED & keys)
    supporting_count = len((_RFC7807_REQUIRED | _RFC7807_SUPPORTING) & keys)
    return has_required and supporting_count >= 2


def extract_rfc7807(body: dict[str, Any]) -> ExtractedErrorInfo:
    message = get_typed_field(body, str, "detail") or get_typed_field(
        body, str, "title"
    )
    error_code = get_typed_field(body, str, "type")
    trace_id = get_typed_field(body, str, "traceId", "instance")

    return ExtractedErrorInfo(
        message=message,
        error_code=error_code,
        trace_id=trace_id,
    )



from unittest.mock import Mock, patch

import pytest
from pytest_httpx import HTTPXMock

from uipath.platform import UiPathApiConfig, UiPathExecutionContext
from uipath.platform.common.constants import HEADER_FOLDER_KEY, HEADER_USER_AGENT
from uipath.platform.orchestrator import McpService
from uipath.platform.orchestrator._folder_service import FolderService
from uipath.platform.orchestrator.mcp import McpServer


@pytest.fixture
def folders_service(
    config: UiPathApiConfig,
    execution_context: UiPathExecutionContext,
    monkeypatch: pytest.MonkeyPatch,
) -> FolderService:
    monkeypatch.setenv("UIPATH_FOLDER_KEY", "test-folder-key")
    return FolderService(config=config, execution_context=execution_context)


@pytest.fixture
def service(
    config: UiPathApiConfig,
    execution_context: UiPathExecutionContext,
    folders_service: FolderService,
    monkeypatch: pytest.MonkeyPatch,
) -> McpService:
    monkeypatch.setenv("UIPATH_FOLDER_KEY", "test-folder-key")
    return McpService(
        config=config,
        execution_context=execution_context,
        folders_service=folders_service,
    )


class TestMcpService:
    class TestListServers:
        def test_list_with_folder_path(
            self,
            httpx_mock: HTTPXMock,
            service: McpService,
            base_url: str,
            org: str,
            tenant: str,
            version: str,
        ) -> None:
            """Test listing MCP servers with a folder_path parameter that gets resolved."""
            mock_servers = [
                {
                    "id": "server-id-1",
                    "name": "Test MCP Server",
                    "slug": "test-mcp-server",
                    "description": "Test description",
                    "version": "1.0.0",
                    "createdAt": "2025-07-24T11:30:52.031427",
                    "updatedAt": "2025-07-24T12:29:53.4765887",
                    "isActive": True,
                    "type": 2,
                    "status": 1,
                    "command": "",
                    "arguments": "",
                    "environmentVariables": "",
                    "processKey": "test-process-key",
                    "folderKey": "test-folder-key",
                    "runtimesCount": 0,
                    "mcpUrl": "https://test.com/mcp/test-mcp-server",
                }
            ]

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/orchestrator_/api/FoldersNavigation/GetFoldersForCurrentUser?searchText=test-folder-path&skip=0&take=20",
                status_code=200,
                json={
                    "PageItems": [
                        {
                            "Key": "resolved-folder-key",
                            "FullyQualifiedName": "test-folder-path",
                        }
                    ],
                },
            )

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/agenthub_/api/servers",
                status_code=200,
                json=mock_servers,
            )

            servers = service.list(folder_path="test-folder-path")

            assert len(servers) == 1
            assert isinstance(servers[0], McpServer)
            assert servers[0].name == "Test MCP Server"

            requests = httpx_mock.get_requests()
            assert len(requests) == 2

            servers_request = requests[1]
            assert servers_request.method == "GET"
            assert (
                servers_request.url == f"{base_url}{org}{tenant}/agenthub_/api/servers"
            )
            assert HEADER_FOLDER_KEY in servers_request.headers
            assert servers_request.headers[HEADER_FOLDER_KEY] == "resolved-folder-key"

        def test_list_without_folder_raises_error(
            self,
            config: UiPathApiConfig,
            execution_context: UiPathExecutionContext,
            monkeypatch: pytest.MonkeyPatch,
        ) -> None:
            """Test that listing servers without a folder_path raises ValueError."""
            monkeypatch.delenv("UIPATH_FOLDER_KEY", raising=False)
            monkeypatch.delenv("UIPATH_FOLDER_PATH", raising=False)

            folders_service = FolderService(
                config=config, execution_context=execution_context
            )
            service = McpService(
                config=config,
                execution_context=execution_context,
                folders_service=folders_service,
            )

            with pytest.raises(
                ValueError,
                match="Cannot obtain folder_key without providing folder_path",
            ):
                service.list()

        @pytest.mark.anyio
        async def test_list_async(
            self,
            httpx_mock: HTTPXMock,
            service: McpService,
            base_url: str,
            org: str,
            tenant: str,
            version: str,
        ) -> None:
            """Test asynchronously listing MCP servers."""
            mock_servers = [
                {
                    "id": "server-id-1",
                    "name": "Async Test Server",
                    "slug": "async-test-server",
                    "description": "Async test description",
                    "version": "1.0.0",
                    "createdAt": "2025-07-24T11:30:52.031427",
                    "updatedAt": "2025-07-24T12:29:53.4765887",
                    "isActive": True,
                    "type": 2,
                    "status": 1,
                    "command": "",
                    "arguments": "",
                    "environmentVariables": "",
                    "processKey": "test-process-key",
                    "folderKey": "test-folder-key",
                    "runtimesCount": 0,
                    "mcpUrl": "https://test.com/mcp/async-test-server",
                }
            ]

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/orchestrator_/api/FoldersNavigation/GetFoldersForCurrentUser?searchText=test-folder-path&skip=0&take=20",
                status_code=200,
                json={
                    "PageItems": [
                        {
                            "Key": "test-folder-key",
                            "FullyQualifiedName": "test-folder-path",
                        }
                    ],
                },
            )

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/agenthub_/api/servers",
                status_code=200,
                json=mock_servers,
            )

            servers = await service.list_async(folder_path="test-folder-path")

            assert len(servers) == 1
            assert isinstance(servers[0], McpServer)
            assert servers[0].name == "Async Test Server"

            requests = httpx_mock.get_requests()
            assert len(requests) == 2

            servers_request = requests[1]
            assert servers_request.method == "GET"
            assert (
                servers_request.url == f"{base_url}{org}{tenant}/agenthub_/api/servers"
            )
            assert HEADER_FOLDER_KEY in servers_request.headers
            assert servers_request.headers[HEADER_FOLDER_KEY] == "test-folder-key"
            assert HEADER_USER_AGENT in servers_request.headers
            assert (
                servers_request.headers[HEADER_USER_AGENT]
                == f"UiPath.Python.Sdk/UiPath.Python.Sdk.Activities.McpService.list_async/{version}"
            )

    class TestRetrieveServer:
        def test_retrieve_server_with_folder_path(
            self,
            httpx_mock: HTTPXMock,
            service: McpService,
            base_url: str,
            org: str,
            tenant: str,
            version: str,
        ) -> None:
            """Test retrieving a specific MCP server by slug with folder_path."""
            mock_server = {
                "id": "server-id-1",
                "name": "Test MCP Server",
                "slug": "test-mcp-server",
                "description": "A test server",
                "version": "1.0.0",
                "createdAt": "2025-07-24T11:30:52.031427",
                "updatedAt": "2025-07-24T12:29:53.4765887",
                "isActive": True,
                "type": 2,
                "status": 1,
                "command": "",
                "arguments": "",
                "environmentVariables": "",
                "processKey": "test-process-key",
                "folderKey": "test-folder-key",
                "runtimesCount": 0,
                "mcpUrl": "https://test.com/mcp/test-mcp-server",
            }

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/orchestrator_/api/FoldersNavigation/GetFoldersForCurrentUser?searchText=test-folder-path&skip=0&take=20",
                status_code=200,
                json={
                    "PageItems": [
                        {
                            "Key": "test-folder-key",
                            "FullyQualifiedName": "test-folder-path",
                        }
                    ],
                },
            )

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/agenthub_/api/servers/test-mcp-server",
                status_code=200,
                json=mock_server,
            )

            server = service.retrieve("test-mcp-server", folder_path="test-folder-path")

            assert isinstance(server, McpServer)
            assert server.name == "Test MCP Server"
            assert server.slug == "test-mcp-server"

            requests = httpx_mock.get_requests()
            assert len(requests) == 2

            retrieve_request = requests[1]
            assert retrieve_request.method == "GET"
            assert (
                retrieve_request.url
                == f"{base_url}{org}{tenant}/agenthub_/api/servers/test-mcp-server"
            )
            assert HEADER_FOLDER_KEY in retrieve_request.headers
            assert retrieve_request.headers[HEADER_FOLDER_KEY] == "test-folder-key"
            assert HEADER_USER_AGENT in retrieve_request.headers
            assert (
                retrieve_request.headers[HEADER_USER_AGENT]
                == f"UiPath.Python.Sdk/UiPath.Python.Sdk.Activities.McpService.retrieve/{version}"
            )

        @pytest.mark.anyio
        async def test_retrieve_server_async(
            self,
            httpx_mock: HTTPXMock,
            service: McpService,
            base_url: str,
            org: str,
            tenant: str,
            version: str,
        ) -> None:
            """Test asynchronously retrieving a specific MCP server."""
            mock_server = {
                "id": "server-id-1",
                "name": "Async Test Server",
                "slug": "async-test-server",
                "description": "Async test server",
                "version": "1.0.0",
                "createdAt": "2025-07-24T11:30:52.031427",
                "updatedAt": "2025-07-24T12:29:53.4765887",
                "isActive": True,
                "type": 2,
                "status": 1,
                "command": "",
                "arguments": "",
                "environmentVariables": "",
                "processKey": "test-process-key",
                "folderKey": "test-folder-key",
                "runtimesCount": 0,
                "mcpUrl": "https://test.com/mcp/async-test-server",
            }

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/orchestrator_/api/FoldersNavigation/GetFoldersForCurrentUser?searchText=test-folder-path&skip=0&take=20",
                status_code=200,
                json={
                    "PageItems": [
                        {
                            "Key": "test-folder-key",
                            "FullyQualifiedName": "test-folder-path",
                        }
                    ],
                },
            )

            httpx_mock.add_response(
                url=f"{base_url}{org}{tenant}/agenthub_/api/servers/async-test-server",
                status_code=200,
                json=mock_server,
            )

            server = await service.retrieve_async(
                "async-test-server", folder_path="test-folder-path"
            )

            assert isinstance(server, McpServer)
            assert server.name == "Async Test Server"

            requests = httpx_mock.get_requests()
            assert len(requests) == 2

            retrieve_request = requests[1]
            assert retrieve_request.method == "GET"
            assert (
                retrieve_request.url
                == f"{base_url}{org}{tenant}/agenthub_/api/servers/async-test-server"
            )
            assert HEADER_FOLDER_KEY in retrieve_request.headers
            assert retrieve_request.headers[HEADER_FOLDER_KEY] == "test-folder-key"
            assert HEADER_USER_AGENT in retrieve_request.headers
            assert (
                retrieve_request.headers[HEADER_USER_AGENT]
                == f"UiPath.Python.Sdk/UiPath.Python.Sdk.Activities.McpService.retrieve_async/{version}"
            )

    class TestRequestKwargs:
        """Test that all methods pass the correct kwargs to request/request_async."""

        def test_list_passes_all_kwargs(self, service: McpService) -> None:
            """Test that list passes all kwargs to request."""
            mock_response = Mock()
            mock_response.json.return_value = [
                {
                    "id": "test-id",
                    "name": "Test Server",
                    "slug": "test-server",
                    "description": "Test",
                    "version": "1.0.0",
                    "createdAt": "2025-07-24T11:30:52.031427",
                    "updatedAt": "2025-07-24T12:29:53.4765887",
                    "isActive": True,
                    "type": 2,
                    "status": 1,
                    "processKey": "test-process-key",
                    "folderKey": "test-folder-key",
                    "mcpUrl": "https://test.com/mcp/test",
                }
            ]

            with patch.object(
                service._folders_service,
                "retrieve_folder_key",
                return_value="test-folder-key",
            ):
                with patch.object(
                    service, "request", return_value=mock_response
                ) as mock_request:
                    service.list(folder_path="test-folder-path")

                    mock_request.assert_called_once()
                    call_kwargs = mock_request.call_args

                    assert "url" in call_kwargs.kwargs
                    assert "params" in call_kwargs.kwargs
                    assert "headers" in call_kwargs.kwargs

                    assert call_kwargs.args[0] == "GET"

                    assert HEADER_FOLDER_KEY in call_kwargs.kwargs["headers"]
                    assert (
                        call_kwargs.kwargs["headers"][HEADER_FOLDER_KEY]
                        == "test-folder-key"
                    )

        @pytest.mark.anyio
        async def test_list_async_passes_all_kwargs(self, service: McpService) -> None:
            """Test that list_async passes all kwargs to request_async."""
            mock_response = Mock()
            mock_response.json.return_value = [
                {
                    "id": "test-id",
                    "name": "Test Server",
                    "slug": "test-server",
                    "description": "Test",
                    "version": "1.0.0",
                    "createdAt": "2025-07-24T11:30:52.031427",
                    "updatedAt": "2025-07-24T12:29:53.4765887",
                    "isActive": True,
                    "type": 2,
                    "status": 1,
                    "processKey": "test-process-key",
                    "folderKey": "test-folder-key",
                    "mcpUrl": "https://test.com/mcp/test",
                }
            ]

            with patch.object(
                service._folders_service,
                "retrieve_folder_key",
                return_value="test-folder-key",
            ):
                with patch.object(
                    service, "request_async", return_value=mock_response
                ) as mock_request:
                    await service.list_async(folder_path="test-folder-path")

                    mock_request.assert_called_once()
                    call_kwargs = mock_request.call_args

                    assert "url" in call_kwargs.kwargs
                    assert "params" in call_kwargs.kwargs
                    assert "headers" in call_kwargs.kwargs

                    assert call_kwargs.args[0] == "GET"

                    assert HEADER_FOLDER_KEY in call_kwargs.kwargs["headers"]
                    assert (
                        call_kwargs.kwargs["headers"][HEADER_FOLDER_KEY]
                        == "test-folder-key"
                    )

        def test_retrieve_passes_all_kwargs(self, service: McpService) -> None:
            """Test that retrieve passes all kwargs to request."""
            mock_response = Mock()
            mock_response.json.return_value = {
                "id": "test-id",
                "name": "Test Server",
                "slug": "test-server",
                "description": "Test",
                "version": "1.0.0",
                "createdAt": "2025-07-24T11:30:52.031427",
                "updatedAt": "2025-07-24T12:29:53.4765887",
                "isActive": True,
                "type": 2,
                "status": 1,
                "processKey": "test-process-key",
                "folderKey": "test-folder-key",
                "mcpUrl": "https://test.com/mcp/test",
            }

            with patch.object(
                service._folders_service,
                "retrieve_folder_key",
                return_value="test-folder-key",
            ):
                with patch.object(
                    service, "request", return_value=mock_response
                ) as mock_request:
                    service.retrieve("test-server", folder_path="test-folder-path")

                    mock_request.assert_called_once()
                    call_kwargs = mock_request.call_args

                    assert "url" in call_kwargs.kwargs
                    assert "params" in call_kwargs.kwargs
                    assert "headers" in call_kwargs.kwargs

                    assert call_kwargs.args[0] == "GET"

                    assert HEADER_FOLDER_KEY in call_kwargs.kwargs["headers"]
                    assert (
                        call_kwargs.kwargs["headers"][HEADER_FOLDER_KEY]
                        == "test-folder-key"
                    )

        @pytest.mark.anyio
        async def test_retrieve_async_passes_all_kwargs(
            self, service: McpService
        ) -> None:
            """Test that retrieve_async passes all kwargs to request_async."""
            mock_response = Mock()
            mock_response.json.return_value = {
                "id": "test-id",
                "name": "Test Server",
                "slug": "test-server",
                "description": "Test",
                "version": "1.0.0",
                "createdAt": "2025-07-24T11:30:52.031427",
                "updatedAt": "2025-07-24T12:29:53.4765887",
                "isActive": True,
                "type": 2,
                "status": 1,
                "processKey": "test-process-key",
                "folderKey": "test-folder-key",
                "mcpUrl": "https://test.com/mcp/test",
            }

            with patch.object(
                service._folders_service,
                "retrieve_folder_key",
                return_value="test-folder-key",
            ):
                with patch.object(
                    service, "request_async", return_value=mock_response
                ) as mock_request:
                    await service.retrieve_async(
                        "test-server", folder_path="test-folder-path"
                    )

                    mock_request.assert_called_once()
                    call_kwargs = mock_request.call_args

                    assert "url" in call_kwargs.kwargs
                    assert "params" in call_kwargs.kwargs
                    assert "headers" in call_kwargs.kwargs

                    assert call_kwargs.args[0] == "GET"

                    assert HEADER_FOLDER_KEY in call_kwargs.kwargs["headers"]
                    assert (
                        call_kwargs.kwargs["headers"][HEADER_FOLDER_KEY]
                        == "test-folder-key"
                    )



import json
import os
import uuid
from typing import Any, Dict, Optional, Union

from opentelemetry import trace
from opentelemetry.trace import format_span_id
from uipath.core.tracing import traced

from ..attachments import Attachment
from ..common._base_service import BaseService
from ..common._bindings import resource_override
from ..common._config import UiPathApiConfig, UiPathConfig
from ..common._execution_context import UiPathExecutionContext
from ..common._folder_context import FolderContext, header_folder
from ..common._models import Endpoint, RequestSpec
from ..common._span_utils import _SpanUtils
from ..common.constants import ENV_JOB_KEY, HEADER_JOB_KEY
from ._attachments_service import AttachmentsService
from .job import Job


class ProcessesService(FolderContext, BaseService):
    """Service for managing and executing UiPath automation processes.

    Processes (also known as automations or workflows) are the core units of
    automation in UiPath, representing sequences of activities that perform
    specific business tasks.
    """

    _INPUT_ARGUMENTS_SIZE_LIMIT = 10000

    def __init__(
        self,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        attachment_service: AttachmentsService,
    ) -> None:
        self._attachments_service = attachment_service
        super().__init__(config=config, execution_context=execution_context)

    @resource_override(resource_type="process")
    @traced(name="processes_invoke", run_type="uipath")
    def invoke(
        self,
        name: str,
        input_arguments: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
        attachments: Optional[list[Attachment]] = None,
        parent_operation_id: Optional[str] = None,
        **kwargs: Any,
    ) -> Job:
        """Start execution of a process by its name.

        Related Activity: [Invoke Process](https://docs.uipath.com/activities/other/latest/workflow/invoke-process)

        Args:
            name (str): The name of the process to execute.
            input_arguments (Optional[Dict[str, Any]]): The input arguments to pass to the process.
            attachments (Optional[list]): List of Attachment objects to pass to the process.
            folder_key (Optional[str]): The key of the folder to execute the process in. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder to execute the process in. Override the default one set in the SDK config.
            parent_operation_id (Optional[str]): The parent operation ID for BTS tracking correlation.

        Returns:
            Job: The job execution details.

        Examples:
            ```python
            from uipath.platform import UiPath

            client = UiPath()

            client.processes.invoke(name="MyProcess")
            ```

            ```python
            # if you want to execute the process in a specific folder
            # another one than the one set in the SDK config
            from uipath.platform import UiPath

            client = UiPath()

            client.processes.invoke(name="MyProcess", folder_path="my-folder-key")
            ```
        """
        input_data = self._handle_input_arguments(
            input_arguments=input_arguments,
            attachments=attachments,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        spec = self._invoke_spec(
            name,
            input_data=input_data,
            folder_key=folder_key,
            folder_path=folder_path,
            parent_span_id=kwargs.get("parent_span_id"),
            parent_operation_id=parent_operation_id,
        )
        response = self.request(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            content=spec.content,
            headers=spec.headers,
        )

        return Job.model_validate(response.json()["value"][0])

    @resource_override(resource_type="process")
    @traced(name="processes_invoke", run_type="uipath")
    async def invoke_async(
        self,
        name: str,
        input_arguments: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
        attachments: Optional[list[Attachment]] = None,
        parent_operation_id: Optional[str] = None,
        **kwargs: Any,
    ) -> Job:
        """Asynchronously start execution of a process by its name.

        Related Activity: [Invoke Process](https://docs.uipath.com/activities/other/latest/workflow/invoke-process)

        Args:
            name (str): The name of the process to execute.
            input_arguments (Optional[Dict[str, Any]]): The input arguments to pass to the process.
            attachments (Optional[list]): List of Attachment objects to pass to the process.
            folder_key (Optional[str]): The key of the folder to execute the process in. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder to execute the process in. Override the default one set in the SDK config.
            parent_operation_id (Optional[str]): The parent operation ID for BTS tracking correlation.

        Returns:
            Job: The job execution details.

        Examples:
            ```python
            import asyncio

            from uipath.platform import UiPath

            sdk = UiPath()

            async def main():
                job = await sdk.processes.invoke_async("testAppAction")
                print(job)

            asyncio.run(main())
            ```
        """
        input_data = await self._handle_input_arguments_async(
            input_arguments=input_arguments,
            attachments=attachments,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        spec = self._invoke_spec(
            name,
            input_data=input_data,
            folder_key=folder_key,
            folder_path=folder_path,
            parent_span_id=kwargs.get("parent_span_id"),
            parent_operation_id=parent_operation_id,
        )

        response = await self.request_async(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            content=spec.content,
            headers=spec.headers,
        )

        return Job.model_validate(response.json()["value"][0])

    @property
    def custom_headers(self) -> Dict[str, str]:
        return self.folder_headers

    @staticmethod
    def _prepare_link_attachments(
        attachments: Optional[list[Attachment]],
    ) -> Optional[list[Dict[str, str]]]:
        """Format attachments for process invocation payload."""
        if not attachments:
            return None

        link_attachments = [
            {"attachmentId": str(att.id)} for att in attachments if att.id is not None
        ]
        return link_attachments if link_attachments else None

    def _handle_input_arguments(
        self,
        input_arguments: Optional[Dict[str, Any]] = None,
        attachments: Optional[list[Attachment]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Handle input arguments and attachments, storing as attachment if they exceed size limit.

        Args:
            input_arguments: The input arguments to process
            attachments: List of Attachment objects to pass to the process
            folder_key: The folder key for attachment storage
            folder_path: The folder path for attachment storage

        Returns:
            Dict containing either "InputArguments" or "InputFile" key, and optionally "Attachments"
        """
        result: Dict[str, Any] = {}

        # handle input arguments
        if not input_arguments:
            result["InputArguments"] = json.dumps({})
        else:
            # If payload exceeds limit, store as attachment
            payload_json = json.dumps(input_arguments)
            if len(payload_json) > self._INPUT_ARGUMENTS_SIZE_LIMIT:
                attachment_id = self._attachments_service.upload(
                    name=f"{uuid.uuid4()}.json",
                    content=payload_json,
                    folder_key=folder_key,
                    folder_path=folder_path,
                )
                result["InputFile"] = str(attachment_id)
            else:
                result["InputArguments"] = payload_json

        link_attachments = self._prepare_link_attachments(attachments)
        if link_attachments:
            result["Attachments"] = link_attachments

        return result

    async def _handle_input_arguments_async(
        self,
        input_arguments: Optional[Dict[str, Any]] = None,
        attachments: Optional[list[Attachment]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Handle input arguments and attachments, storing as attachment if they exceed size limit.

        Args:
            input_arguments: The input arguments to process
            attachments: List of Attachment objects to pass to the process
            folder_key: The folder key for attachment storage
            folder_path: The folder path for attachment storage

        Returns:
            Dict containing either "InputArguments" or "InputFile" key, and optionally "Attachments"
        """
        result: Dict[str, Any] = {}

        if not input_arguments:
            result["InputArguments"] = json.dumps({})
        else:
            payload_json = json.dumps(input_arguments)
            if len(payload_json) > self._INPUT_ARGUMENTS_SIZE_LIMIT:
                attachment_id = await self._attachments_service.upload_async(
                    name=f"{uuid.uuid4()}.json",
                    content=payload_json,
                    folder_key=folder_key,
                    folder_path=folder_path,
                )
                result["InputFile"] = str(attachment_id)
            else:
                result["InputArguments"] = payload_json

        formatted_attachments = self._prepare_link_attachments(attachments)
        if formatted_attachments:
            result["Attachments"] = formatted_attachments

        return result

    @staticmethod
    def _add_tracing(
        payload: Dict[str, Any],
        trace_id: Optional[str] = None,
        parent_span_id: Optional[Union[str, int]] = None,
    ) -> None:
        """Enrich payload with trace context for cross-process correlation."""
        if not trace_id:
            return

        payload["TraceId"] = _SpanUtils.normalize_trace_id(trace_id)
        if not parent_span_id:
            span_context = trace.get_current_span().get_span_context()
            if span_context.span_id:
                parent_span_id = format_span_id(span_context.span_id)
        if parent_span_id:
            if isinstance(parent_span_id, int):
                parent_span_id = format_span_id(parent_span_id)
            payload["ParentSpanId"] = _SpanUtils.normalize_span_id(parent_span_id)

    def _invoke_spec(
        self,
        name: str,
        input_data: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
        parent_span_id: Optional[str] = None,
        parent_operation_id: Optional[str] = None,
    ) -> RequestSpec:
        payload: Dict[str, Any] = {"ReleaseName": name, **(input_data or {})}
        self._add_tracing(payload, UiPathConfig.trace_id, parent_span_id)

        if parent_operation_id:
            payload["ParentOperationId"] = parent_operation_id

        request_spec = RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.StartJobs"
            ),
            json={"startInfo": payload},
            headers={
                **header_folder(folder_key, folder_path),
            },
        )
        job_key = os.environ.get(ENV_JOB_KEY, None)
        if job_key:
            request_spec.headers[HEADER_JOB_KEY] = job_key

        return request_spec



"""UiPath Automation Tracker (BTS) Models.

This module contains models and service for tracking business transactions
and operations via the Business Transaction Service (BTS).
"""

from ._automation_tracker_service import AutomationTrackerService
from .automation_tracker import (
    OperationPayload,
    OperationStatus,
    TransactionPayload,
    TransactionStatus,
)

__all__ = [
    "AutomationTrackerService",
    "OperationPayload",
    "OperationStatus",
    "TransactionPayload",
    "TransactionStatus",
]



import inspect
import json
import logging
import os
from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum
from os import environ as env
from typing import Any, Dict, List, Optional

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.trace import StatusCode
from pydantic import BaseModel, ConfigDict, Field
from uipath.core.serialization import serialize_json

logger = logging.getLogger(__name__)

# SourceEnum.Robots = 4 (default for Python SDK / coded agents)
DEFAULT_SOURCE = 4


class AttachmentProvider(IntEnum):
    ORCHESTRATOR = 0


class AttachmentDirection(IntEnum):
    NONE = 0
    IN = 1
    OUT = 2


class SpanAttachment(BaseModel):
    """Represents an attachment in the UiPath tracing system."""

    model_config = ConfigDict(populate_by_name=True, use_enum_values=True)

    id: str = Field(..., alias="id")
    file_name: str = Field(..., alias="fileName")
    mime_type: str = Field(..., alias="mimeType")
    provider: AttachmentProvider = Field(
        default=AttachmentProvider.ORCHESTRATOR, alias="provider"
    )
    direction: AttachmentDirection = Field(
        default=AttachmentDirection.NONE, alias="direction"
    )


@dataclass
class UiPathSpan:
    """Represents a span in the UiPath tracing system.

    Note: attributes can be either a JSON string (backwards compatible) or a dict (optimized).
    IDs are stored as OTEL hex strings (32 chars for trace_id, 16 chars for span_id/parent_id).
    """

    id: str  # 16-char hex (OTEL span ID format)
    trace_id: str  # 32-char hex (OTEL trace ID format)
    name: str
    attributes: str | Dict[str, Any]  # Support both str (legacy) and dict (optimized)
    parent_id: Optional[str] = None  # 16-char hex (OTEL span ID format)
    start_time: str = field(default_factory=lambda: datetime.now().isoformat())
    end_time: str = field(default_factory=lambda: datetime.now().isoformat())
    status: int = 1
    created_at: str = field(default_factory=lambda: datetime.now().isoformat() + "Z")
    updated_at: str = field(default_factory=lambda: datetime.now().isoformat() + "Z")
    organization_id: Optional[str] = field(
        default_factory=lambda: env.get("UIPATH_ORGANIZATION_ID", "")
    )
    tenant_id: Optional[str] = field(
        default_factory=lambda: env.get("UIPATH_TENANT_ID", "")
    )
    expiry_time_utc: Optional[str] = None
    folder_key: Optional[str] = field(
        default_factory=lambda: env.get("UIPATH_FOLDER_KEY", "")
    )
    source: int = DEFAULT_SOURCE
    span_type: str = "Coded Agents"
    process_key: Optional[str] = field(
        default_factory=lambda: env.get("UIPATH_PROCESS_UUID")
    )
    reference_id: Optional[str] = field(
        default_factory=lambda: env.get("TRACE_REFERENCE_ID")
    )

    job_key: Optional[str] = field(default_factory=lambda: env.get("UIPATH_JOB_KEY"))

    # Top-level fields for internal tracing schema
    execution_type: Optional[int] = None
    agent_version: Optional[str] = None
    attachments: Optional[List[SpanAttachment]] = None

    def to_dict(self, serialize_attributes: bool = True) -> Dict[str, Any]:
        """Convert the Span to a dictionary suitable for JSON serialization.

        Args:
            serialize_attributes: If True and attributes is a dict, serialize to JSON string.
                                 If False, keep attributes as-is (dict or str).
                                 Default True for backwards compatibility.
        """
        attributes_out = self.attributes
        if serialize_attributes and isinstance(self.attributes, dict):
            attributes_out = json.dumps(self.attributes)

        attachments_out = None
        if self.attachments is not None:
            attachments_out = [
                {
                    "Id": att.id,
                    "FileName": att.file_name,
                    "MimeType": att.mime_type,
                    "Provider": int(att.provider),
                    "Direction": int(att.direction),
                }
                for att in self.attachments
            ]

        return {
            "Id": self.id,
            "TraceId": self.trace_id,
            "ParentId": self.parent_id,
            "Name": self.name,
            "StartTime": self.start_time,
            "EndTime": self.end_time,
            "Attributes": attributes_out,
            "Status": self.status,
            "CreatedAt": self.created_at,
            "UpdatedAt": self.updated_at,
            "OrganizationId": self.organization_id,
            "TenantId": self.tenant_id,
            "ExpiryTimeUtc": self.expiry_time_utc,
            "FolderKey": self.folder_key,
            "Source": self.source,
            "SpanType": self.span_type,
            "ProcessKey": self.process_key,
            "JobKey": self.job_key,
            "ReferenceId": self.reference_id,
            "ExecutionType": self.execution_type,
            "AgentVersion": self.agent_version,
            "Attachments": attachments_out,
        }


class _SpanUtils:
    @staticmethod
    def normalize_trace_id(value: str) -> str:
        """Normalize trace ID to 32-char OTEL hex format.

        Accepts both UUID format (with dashes) and OTEL hex format (32 chars).
        Returns lowercase 32-char hex string.
        """
        # Remove dashes if UUID format
        normalized = value.replace("-", "").lower()
        if len(normalized) != 32:
            raise ValueError(f"Invalid trace ID format: {value}")
        return normalized

    @staticmethod
    def normalize_span_id(value: str) -> str:
        """Normalize span ID to 16-char OTEL hex format.

        Accepts both UUID format (with dashes, uses last 16 hex chars) and OTEL hex format (16 chars).
        Returns lowercase 16-char hex string.
        """
        # Remove dashes if UUID format
        normalized = value.replace("-", "").lower()
        if len(normalized) == 32:
            # UUID format - take last 16 chars (span ID portion)
            return normalized[16:]
        elif len(normalized) == 16:
            return normalized
        else:
            raise ValueError(f"Invalid span ID format: {value}")

    @staticmethod
    def otel_span_to_uipath_span(
        otel_span: ReadableSpan,
        custom_trace_id: Optional[str] = None,
        serialize_attributes: bool = True,
    ) -> UiPathSpan:
        """Convert an OpenTelemetry span to a UiPathSpan.

        Args:
            otel_span: The OpenTelemetry span to convert
            custom_trace_id: Optional custom trace ID to use (UUID or OTEL hex format)
            serialize_attributes: If True, serialize attributes to JSON string (backwards compatible).
                                 If False, keep as dict for optimized processing. Default True.
        """
        # Extract the context information from the OTel span
        span_context = otel_span.get_span_context()

        # Convert to OTEL hex format (32 chars for trace_id, 16 chars for span_id)
        trace_id = format(span_context.trace_id, "032x")
        span_id = format(span_context.span_id, "016x")

        # Override trace_id if custom or env var provided (supports both UUID and hex format)
        trace_id_override = custom_trace_id or os.environ.get("UIPATH_TRACE_ID")
        if trace_id_override:
            trace_id = _SpanUtils.normalize_trace_id(trace_id_override)

        # Get parent span ID if it exists
        parent_id: Optional[str] = None
        if otel_span.parent is not None:
            parent_id = format(otel_span.parent.span_id, "016x")
        else:
            # Only set UIPATH_PARENT_SPAN_ID for root spans (spans without a parent)
            parent_span_id_str = env.get("UIPATH_PARENT_SPAN_ID")
            if parent_span_id_str:
                parent_id = _SpanUtils.normalize_span_id(parent_span_id_str)

        # Build attributes dict efficiently
        # Use the otel attributes as base - we only add new keys, don't modify existing
        otel_attrs = otel_span.attributes if otel_span.attributes else {}
        # Only copy if we need to modify - we'll build attributes_dict lazily
        attributes_dict: dict[str, Any] = dict(otel_attrs) if otel_attrs else {}

        # Map status
        status = 1  # Default to OK
        if otel_span.status.status_code == StatusCode.ERROR:
            status = 2  # Error
            attributes_dict["error"] = otel_span.status.description

        # Process inputs - avoid redundant parsing if already parsed
        original_inputs = otel_attrs.get("input", None)
        if original_inputs:
            if isinstance(original_inputs, str):
                try:
                    attributes_dict["input.value"] = json.loads(original_inputs)
                    attributes_dict["input.mime_type"] = "application/json"
                except Exception:
                    attributes_dict["input.value"] = original_inputs
            else:
                attributes_dict["input.value"] = original_inputs

        # Process outputs - avoid redundant parsing if already parsed
        original_outputs = otel_attrs.get("output", None)
        if original_outputs:
            if isinstance(original_outputs, str):
                try:
                    attributes_dict["output.value"] = json.loads(original_outputs)
                    attributes_dict["output.mime_type"] = "application/json"
                except Exception:
                    attributes_dict["output.value"] = original_outputs
            else:
                attributes_dict["output.value"] = original_outputs

        # Add events as additional attributes if they exist
        if otel_span.events:
            events_list = [
                {
                    "name": event.name,
                    "timestamp": event.timestamp,
                    "attributes": dict(event.attributes) if event.attributes else {},
                }
                for event in otel_span.events
            ]
            attributes_dict["events"] = events_list

        # Add links as additional attributes if they exist
        if hasattr(otel_span, "links") and otel_span.links:
            links_list = [
                {
                    "trace_id": link.context.trace_id,
                    "span_id": link.context.span_id,
                    "attributes": dict(link.attributes) if link.attributes else {},
                }
                for link in otel_span.links
            ]
            attributes_dict["links"] = links_list

        span_type_value = attributes_dict.get("span_type", "OpenTelemetry")
        span_type = str(span_type_value)

        # Top-level fields for internal tracing schema
        execution_type = attributes_dict.get("executionType")
        agent_version = attributes_dict.get("agentVersion")
        reference_id = attributes_dict.get("referenceId")

        # Source: override via uipath.source attribute, else DEFAULT_SOURCE
        uipath_source = attributes_dict.get("uipath.source")
        source = uipath_source if isinstance(uipath_source, int) else DEFAULT_SOURCE

        attachments = None
        attachments_data = attributes_dict.get("attachments")
        if attachments_data:
            try:
                attachments_list = json.loads(attachments_data)
                attachments = [
                    SpanAttachment(
                        id=att.get("id"),
                        file_name=att.get("fileName", ""),
                        mime_type=att.get("mimeType", ""),
                        provider=att.get("provider", 0),
                        direction=att.get("direction", 0),
                    )
                    for att in attachments_list
                ]
            except Exception as e:
                logger.warning(f"Error processing attachments: {e}")

        # Create UiPathSpan from OpenTelemetry span
        start_time = datetime.fromtimestamp(
            (otel_span.start_time or 0) / 1e9
        ).isoformat()

        end_time_str = None
        if otel_span.end_time is not None:
            end_time_str = datetime.fromtimestamp(
                (otel_span.end_time or 0) / 1e9
            ).isoformat()
        else:
            end_time_str = datetime.now().isoformat()

        return UiPathSpan(
            id=span_id,
            trace_id=trace_id,
            parent_id=parent_id,
            name=otel_span.name,
            attributes=json.dumps(attributes_dict)
            if serialize_attributes
            else attributes_dict,
            start_time=start_time,
            end_time=end_time_str,
            status=status,
            span_type=span_type,
            execution_type=execution_type,
            agent_version=agent_version,
            reference_id=reference_id,
            source=source,
            attachments=attachments,
        )

    @staticmethod
    def format_object_for_trace_json(
        input_object: Any,
    ) -> str:
        """Return a JSON string of inputs from the function signature."""
        return serialize_json(input_object)

    @staticmethod
    def format_args_for_trace(
        signature: inspect.Signature, *args: Any, **kwargs: Any
    ) -> Dict[str, Any]:
        try:
            """Return a dictionary of inputs from the function signature."""
            # Create a parameter mapping by partially binding the arguments

            parameter_binding = signature.bind_partial(*args, **kwargs)

            # Fill in default values for any unspecified parameters
            parameter_binding.apply_defaults()

            # Extract the input parameters, skipping special Python parameters
            result = {}
            for name, value in parameter_binding.arguments.items():
                # Skip class and instance references
                if name in ("self", "cls"):
                    continue

                # Handle **kwargs parameters specially
                param_info = signature.parameters.get(name)
                if param_info and param_info.kind == inspect.Parameter.VAR_KEYWORD:
                    # Flatten nested kwargs directly into the result
                    if isinstance(value, dict):
                        result.update(value)
                else:
                    # Regular parameter
                    result[name] = value

            return result
        except Exception as e:
            logger.warning(
                f"Error formatting arguments for trace: {e}. Using args and kwargs directly."
            )
            return {"args": args, "kwargs": kwargs}

    @staticmethod
    def spans_to_llm_context(spans: list[ReadableSpan]) -> str:
        """Convert spans to a formatted conversation history string suitable for LLM context.

        Includes function calls (including LLM calls) with their inputs and outputs.
        """
        history = []
        for span in spans:
            attributes = dict(span.attributes) if span.attributes else {}

            input_value = attributes.get("input.value")
            output_value = attributes.get("output.value")
            telemetry_filter = attributes.get("telemetry.filter")

            if not input_value or not output_value or telemetry_filter == "drop":
                continue

            history.append(f"Function: {span.name}")
            history.append(f"Input: {input_value}")
            history.append(f"Output: {output_value}")
            history.append("")

        if not history:
            return "(empty)"

        return "\n".join(history)



import httpx
from tenacity import Future, RetryCallState, Retrying

from uipath.platform.common.retry import (
    _MAX_BACKOFF_DELAY,
    _MAX_RETRY_AFTER_DELAY,
    exponential_backoff_with_jitter,
    extract_retry_after_from_chain,
    is_retryable_platform_exception,
    is_retryable_response,
    parse_retry_after,
    platform_wait_strategy,
)
from uipath.platform.errors import EnrichedException


class TestParseRetryAfter:
    def test_valid_integer(self):
        assert parse_retry_after("120") == 120.0

    def test_valid_float(self):
        assert parse_retry_after("1.5") == 1.5

    def test_zero(self):
        assert parse_retry_after("0") == 0.0

    def test_whitespace(self):
        assert parse_retry_after("  30  ") == 30.0

    def test_negative_returns_none(self):
        assert parse_retry_after("-1") is None

    def test_non_numeric_returns_none(self):
        assert parse_retry_after("not-a-number") is None

    def test_http_date_returns_none(self):
        assert parse_retry_after("Wed, 21 Oct 2015 07:28:00 GMT") is None

    def test_empty_string_returns_none(self):
        assert parse_retry_after("") is None


class TestExponentialBackoffWithJitter:
    def test_first_attempt(self):
        result = exponential_backoff_with_jitter(attempt=1, initial=1.0)
        # 1.0 * 2^0 + jitter(0,1) = 1.0 + [0, 1)
        assert 1.0 <= result < 2.0

    def test_second_attempt(self):
        result = exponential_backoff_with_jitter(attempt=2, initial=1.0)
        # 1.0 * 2^1 + jitter(0,1) = 2.0 + [0, 1)
        assert 2.0 <= result < 3.0

    def test_third_attempt(self):
        result = exponential_backoff_with_jitter(attempt=3, initial=1.0)
        # 1.0 * 2^2 + jitter(0,1) = 4.0 + [0, 1)
        assert 4.0 <= result < 5.0

    def test_growth_pattern(self):
        # With large enough initial, growth dominates jitter
        results = [
            exponential_backoff_with_jitter(attempt=i, initial=10.0)
            for i in range(1, 5)
        ]
        # Each base should roughly double (jitter is <=1.0, base starts at 10)
        for i in range(1, len(results)):
            assert results[i] > results[i - 1]

    def test_custom_initial(self):
        result = exponential_backoff_with_jitter(attempt=1, initial=5.0)
        assert 5.0 <= result < 6.0


class TestExtractRetryAfterFromChain:
    @staticmethod
    def _make_http_status_error(
        status_code: int, retry_after: str | None = None
    ) -> httpx.HTTPStatusError:
        headers = {}
        if retry_after is not None:
            headers["retry-after"] = retry_after
        response = httpx.Response(
            status_code=status_code,
            headers=headers,
            request=httpx.Request("GET", "https://example.com"),
        )
        return httpx.HTTPStatusError(
            message=f"{status_code}", request=response.request, response=response
        )

    def test_direct_http_status_error(self):
        err = self._make_http_status_error(429, retry_after="30")
        assert extract_retry_after_from_chain(err) == 30.0

    def test_enriched_exception_wrapping(self):
        http_err = self._make_http_status_error(429, retry_after="60")
        enriched = EnrichedException(http_err)
        enriched.__cause__ = http_err
        assert extract_retry_after_from_chain(enriched) == 60.0

    def test_missing_header(self):
        err = self._make_http_status_error(500)
        assert extract_retry_after_from_chain(err) is None

    def test_unrelated_exception(self):
        err = ValueError("something unrelated")
        assert extract_retry_after_from_chain(err) is None

    def test_zero_retry_after(self):
        err = self._make_http_status_error(429, retry_after="0")
        assert extract_retry_after_from_chain(err) == 0.0

    def test_negative_retry_after_ignored(self):
        err = self._make_http_status_error(429, retry_after="-5")
        assert extract_retry_after_from_chain(err) is None


def _make_http_status_error(
    status_code: int, retry_after: str | None = None
) -> httpx.HTTPStatusError:
    headers = {}
    if retry_after is not None:
        headers["retry-after"] = retry_after
    response = httpx.Response(
        status_code=status_code,
        headers=headers,
        request=httpx.Request("GET", "https://example.com"),
    )
    return httpx.HTTPStatusError(
        message=f"{status_code}", request=response.request, response=response
    )


def _make_retry_state(exception: BaseException, attempt: int = 1) -> RetryCallState:
    rs = RetryCallState(Retrying(), None, None, None)
    rs.attempt_number = attempt
    f = Future(attempt)
    f.set_exception(exception)
    rs.outcome = f
    return rs


class TestIsRetryablePlatformException:
    def test_connect_timeout(self):
        err = httpx.ConnectTimeout("timed out")
        assert is_retryable_platform_exception(err) is True

    def test_timeout_exception(self):
        err = httpx.TimeoutException("timed out")
        assert is_retryable_platform_exception(err) is True

    def test_enriched_408(self):
        http_err = _make_http_status_error(408)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is True

    def test_enriched_429(self):
        http_err = _make_http_status_error(429)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is True

    def test_enriched_502(self):
        http_err = _make_http_status_error(502)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is True

    def test_enriched_503(self):
        http_err = _make_http_status_error(503)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is True

    def test_enriched_504(self):
        http_err = _make_http_status_error(504)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is True

    def test_enriched_524_cloudflare_timeout(self):
        http_err = _make_http_status_error(524)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is True

    def test_enriched_400_not_retryable(self):
        http_err = _make_http_status_error(400)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is False

    def test_enriched_500_not_retryable(self):
        http_err = _make_http_status_error(500)
        err = EnrichedException(http_err)
        assert is_retryable_platform_exception(err) is False

    def test_raw_http_error_not_matched(self):
        err = _make_http_status_error(429)
        assert is_retryable_platform_exception(err) is False

    def test_unrelated_exception_not_retryable(self):
        assert is_retryable_platform_exception(ValueError("nope")) is False


class TestIsRetryableResponse:
    def test_500(self):
        assert is_retryable_response(httpx.Response(500)) is True

    def test_502(self):
        assert is_retryable_response(httpx.Response(502)) is True

    def test_503(self):
        assert is_retryable_response(httpx.Response(503)) is True

    def test_599(self):
        assert is_retryable_response(httpx.Response(599)) is True

    def test_200_not_retryable(self):
        assert is_retryable_response(httpx.Response(200)) is False

    def test_400_not_retryable(self):
        assert is_retryable_response(httpx.Response(400)) is False

    def test_429_not_retryable(self):
        assert is_retryable_response(httpx.Response(429)) is False


class TestPlatformWaitStrategy:
    def test_uses_retry_after_header(self):
        http_err = _make_http_status_error(429, retry_after="5")
        rs = _make_retry_state(http_err)
        assert platform_wait_strategy(rs) == 5.0

    def test_caps_retry_after_at_max(self):
        http_err = _make_http_status_error(429, retry_after="999")
        rs = _make_retry_state(http_err)
        assert platform_wait_strategy(rs) == _MAX_RETRY_AFTER_DELAY

    def test_falls_back_to_backoff_without_header(self):
        http_err = _make_http_status_error(503)
        rs = _make_retry_state(http_err, attempt=1)
        wait = platform_wait_strategy(rs)
        # attempt 1: 1.0 * 2^0 + jitter(0,1) = [1.0, 2.0)
        assert 1.0 <= wait < 2.0

    def test_backoff_capped_at_max(self):
        http_err = _make_http_status_error(503)
        # attempt 5: 1.0 * 2^4 = 16.0 + jitter, but capped at _MAX_BACKOFF_DELAY
        rs = _make_retry_state(http_err, attempt=5)
        assert platform_wait_strategy(rs) == _MAX_BACKOFF_DELAY

    def test_no_outcome_falls_back_to_backoff(self):
        rs = RetryCallState(Retrying(), None, None, None)
        rs.attempt_number = 1
        rs.outcome = None
        wait = platform_wait_strategy(rs)
        assert 1.0 <= wait < 2.0



from dataclasses import dataclass, field
from typing import Any, Optional, Union


class Endpoint(str):
    """A string subclass representing a normalized API endpoint path.

    This class ensures consistent endpoint formatting by:
    - Adding a leading slash if missing
    - Removing trailing slashes (except for root '/')
    - Stripping query parameters

    The class supports string formatting for dynamic path parameters.

    Examples:
        >>> endpoint = Endpoint("/api/v1/users/{id}")
        >>> endpoint.format(id=123)
        '/api/v1/users/123'

        >>> endpoint = Endpoint("projects")
        >>> str(endpoint)
        '/projects'

    Args:
        endpoint (str): The endpoint path to normalize. May include format placeholders
            for dynamic values (e.g. "/users/{id}").

    Raises:
        ValueError: If format() is called with None or empty string arguments.
    """

    def __new__(cls, endpoint: str) -> "Endpoint":
        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"

        if endpoint != "/" and endpoint.endswith("/"):
            endpoint = endpoint[:-1]

        endpoint = endpoint.split("?")[0]

        return super().__new__(cls, endpoint)

    def format(self, *args: Any, **kwargs: Any) -> str:
        """Formats the endpoint with the given arguments."""
        for index, arg in enumerate(args):
            if not self._is_valid_value(arg):
                raise ValueError(f"Positional argument `{index}` is `{arg}`.")

        for key, value in kwargs.items():
            if not self._is_valid_value(value):
                raise ValueError(f"Keyword argument `{key}` is `{value}`.")

        return super().format(*args, **kwargs)

    def __repr__(self) -> str:
        return f"Endpoint({super().__str__()!r})"

    def _is_valid_value(self, value: Any) -> bool:
        return value is not None and value != ""

    @property
    def service(self) -> str:
        """Extracts and returns the service name from the endpoint path.

        The service name is expected to be the first path segment after the leading slash,
        with any underscores removed.

        Examples:
            >>> endpoint = Endpoint("/cloud_/projects")
            >>> endpoint.service
            'cloud'

            >>> endpoint = Endpoint("/automation_hub_/assets")
            >>> endpoint.service
            'automationhub'

        Returns:
            str: The service name with underscores removed.
        """
        return self.split("/")[1].replace("_", "")


@dataclass
class RequestSpec:
    """Encapsulates the configuration for making an HTTP request.

    This class contains all necessary parameters to construct and send an HTTP request,
    including the HTTP method, endpoint, query parameters, headers, and various forms
    of request body data (content, JSON, form data).
    """

    method: str
    endpoint: Endpoint
    params: dict[str, Any] = field(default_factory=dict)
    headers: dict[str, Any] = field(default_factory=dict)
    content: Optional[Any] = None
    json: Optional[Any] = None
    data: Optional[Any] = None
    files: Optional[dict[str, Any]] = None
    timeout: Optional[Union[int, float]] = None



from typing import Any

from httpx import HTTPStatusError
from uipath.core.guardrails import (
    GuardrailValidationResult,
    GuardrailValidationResultType,
)
from uipath.core.tracing import traced

from ..common._base_service import BaseService
from ..common._config import UiPathApiConfig
from ..common._execution_context import UiPathExecutionContext
from ..common._models import Endpoint, RequestSpec
from ..errors import EnrichedException
from .guardrails import BuiltInValidatorGuardrail


class GuardrailsService(BaseService):
    """Service for validating text against UiPath Guardrails.

    This service provides an interface for evaluating built-in guardrails such as:

    - PII detection
    - Prompt injection detection

    Deterministic and custom guardrails are not yet supported.

    !!! info "Version Availability"
        This service is available starting from **uipath** version **2.2.12**.
    """

    def __init__(
        self, config: UiPathApiConfig, execution_context: UiPathExecutionContext
    ) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @staticmethod
    def _parse_result(result_str: str) -> GuardrailValidationResultType:
        """Parse result string from API response to GuardrailValidationResultType.

        Args:
            result_str: The result string from the API response (e.g., "VALIDATION_FAILED").

        Returns:
            GuardrailValidationResultType: The parsed validation result type.
        """
        if not result_str:
            return GuardrailValidationResultType.VALIDATION_FAILED

        # Convert uppercase enum name to enum value
        # API: "VALIDATION_FAILED" -> enum: "validation_failed"
        result_value = result_str.lower()
        try:
            return GuardrailValidationResultType(result_value)
        except ValueError:
            # If direct conversion fails, try by enum name
            try:
                return GuardrailValidationResultType[result_str]
            except KeyError:
                # Fallback to validation_failed if unknown
                return GuardrailValidationResultType.VALIDATION_FAILED

    @traced("evaluate_guardrail", run_type="uipath")
    def evaluate_guardrail(
        self,
        input_data: str | dict[str, Any],
        guardrail: BuiltInValidatorGuardrail,
    ) -> GuardrailValidationResult:
        """Validate input text using the provided guardrail.

        Args:
            input_data: The text or structured data to validate. Dictionaries will be converted to a string before validation.
            guardrail: A guardrail instance used for validation.

        Returns:
            GuardrailValidationResult: The outcome of the guardrail evaluation.
        """
        parameters = [
            param.model_dump(by_alias=True) for param in guardrail.validator_parameters
        ]
        payload = {
            "validator": guardrail.validator_type,
            "input": input_data if isinstance(input_data, str) else str(input_data),
            "parameters": parameters,
        }
        spec = RequestSpec(
            method="POST",
            endpoint=Endpoint("/agentsruntime_/api/execution/guardrails/validate"),
            json=payload,
        )
        try:
            response = self.request(
                spec.method,
                url=spec.endpoint,
                json=spec.json,
                headers=spec.headers,
            )
            response_data = response.json()
        except EnrichedException as e:
            # Handle 403 responses: API returns 403 with valid JSON body for
            # ENTITLEMENTS_MISSING or FEATURE_DISABLED cases
            if e.status_code == 403:
                # Access the original HTTPStatusError to get the full response
                original_error = e.__cause__
                if (
                    isinstance(original_error, HTTPStatusError)
                    and original_error.response
                ):
                    try:
                        response_data = original_error.response.json()
                    except Exception:
                        # If JSON parsing fails, re-raise the original exception
                        raise
                else:
                    # Try to parse from response_content if available
                    try:
                        import json

                        response_data = json.loads(e.response_content)
                    except Exception:
                        raise
            else:
                raise

        result = self._parse_result(response_data.get("result", ""))

        reason = response_data.get("details", "")

        # Prepare model data
        model_data = {
            "result": result.value,
            "reason": reason,
        }

        return GuardrailValidationResult.model_validate(model_data)



"""LLMOps service error payload extractor."""

from typing import Any

from .._enriched_exception import ExtractedErrorInfo
from ._helpers import get_typed_field


def extract_llmops(body: dict[str, Any]) -> ExtractedErrorInfo:
    return ExtractedErrorInfo(
        message=get_typed_field(body, str, "errorMessage"),
        error_code=get_typed_field(body, str, "code"),
        trace_id=get_typed_field(body, str, "requestId"),
    )
